// play.cpp: conceptual audio/video stream decoding example
// Author: Guantao Liu, CECS
// Date: 08/30/2015
// 10/30/2015 RD polished for release 0.2.1
// 09/04/2015 GL use 'long long' (not 'double') to avoid floating-point overflow
//               (FPU exception handling shows high performance variation;
//               moreover, integer arithmetic is more realistic for decoder)
// 08/30/2015 GL initial version based on play.sc

#include <stdio.h>
#include <stdlib.h>
#include "systemc.h"

// frame rates
#define AUDIO_DELAY	sc_time(26.12, SC_MS)
#define VIDEO_DELAY	sc_time(33.33, SC_MS)

// patched frame rates
#define AUDIO_DELAY_P	26120, SC_US
#define VIDEO_DELAY_P	33330, SC_US

// simulated time
#define CLIP_LENGTH	sc_time(10, SC_SEC)

// simulated workload unit
#define AUDIOLOAD	 7500000
#define VIDEOLOAD	10000000

// illustrate frame decoding
//#define PRINT_TIMES
#define PRINT_FRAMES

// queue communication

class i_long_long_sender: virtual public sc_interface
{
  public:
    virtual void send(long long d) = 0;
};

class i_long_long_receiver: virtual public sc_interface
{
  public:
    virtual long long receive(void) = 0;
};

class c_long_long_queue: public sc_channel,
	public i_long_long_sender, public i_long_long_receiver
{
  public:
    c_long_long_queue(sc_module_name name, unsigned long size):
		sc_channel(name), m_size(size)
    {
      wr = 0;
      ws = 0;
      p = 0;
      n = 0;
      buffer = NULL;
    }

    void setup(void)
    {
      if (!buffer)
      {
        unsigned long i;
        long long dummy;

        if (!(buffer = (long long*)malloc(sizeof(long long)*m_size)))
        {
          printf("c_typed_queue\n");
          abort();
        }
        for (i = 0; i < m_size; i++)
        {
          memcpy(&buffer[i], &dummy, sizeof(long long));
        }
      }
    }

    void cleanup(void)
    {
      if (!n)
      {
        free(buffer);
        buffer = NULL;
      }
    }

    long long receive(void)
    {
      long long d;
      while (!n)
      {
        wr ++;
        wait(r);
        wr --;
      }

      if (n <= p)
      {
        d = buffer[p - n];
      }
      else
      {
        d = buffer[p + m_size - n];
      }
      n --;

      if (ws)
      {
        s.notify(SC_ZERO_TIME);
      }

      cleanup();
      return d;
    }

    void send(long long d)
    {
      while (n >= m_size)
      {
        ws ++;
        wait(s);
        ws --;
      }

      setup();

      buffer[p] = d;
      p ++;
      if (p >= m_size)
      {
        p = 0;
      }
      n ++;

      if (wr)
      {
        r.notify(SC_ZERO_TIME);
      }
    }

  private:
    sc_event       r, s;
    unsigned long  wr;
    unsigned long  ws;
    unsigned long  p;
    unsigned long  n;
    long long      *buffer;
    unsigned long  m_size;
};

SC_MODULE(Stimulus)
{
  sc_port<i_long_long_sender> VideoStream;
  sc_port<i_long_long_sender> AudioStreamL;
  sc_port<i_long_long_sender> AudioStreamR;

  void main(void)
  {
    sc_time     t1 = sc_time::from_value(0);
    sc_time     t2 = sc_time::from_value(0);
    long long   video = 42;
    long long   audioLeft = 27;
    long long   audioRight = 15;

    while ((t1 < CLIP_LENGTH) || (t2 < CLIP_LENGTH))
    {
      if (t1 < t2)
      {
        VideoStream->send(video);
        t1 += VIDEO_DELAY;
      }
      else
      {
        AudioStreamL->send(audioLeft);
        AudioStreamR->send(audioRight);
        t2 += AUDIO_DELAY;
      }
    }
    printf("\nDone!\n");
    exit(0);
  }

  SC_CTOR(Stimulus)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(VideoCodec)
{
  sc_port<i_long_long_receiver> p1;
  sc_port<i_long_long_sender>   p2;

  long long d;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < VIDEOLOAD; i ++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    long long inFrame, outFrame;

    d = 1;

    while(1)
    {
      inFrame = p1->receive();
      outFrame = decode(inFrame);
      wait(VIDEO_DELAY_P);
      p2->send(outFrame);
    }
  }

  SC_CTOR(VideoCodec)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(AudioCodec)
{
  sc_port<i_long_long_receiver> p1;
  sc_port<i_long_long_sender>   p2;

  long long d;

  long long decode(long long x)
  {
    int i;

    for (i = 0; i < AUDIOLOAD; i++)
    {
      d *= x;
    }
    return d;
  }

  void main(void)
  {
    long long inFrame, outFrame;

    d = 1;

    while(1)
    {
      inFrame = p1->receive();
      outFrame = decode(inFrame);
      wait(AUDIO_DELAY_P);
      p2->send(outFrame);
    }
  }

  SC_CTOR(AudioCodec)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(DUT)
{
  sc_port<i_long_long_receiver> VideoIn;
  sc_port<i_long_long_receiver> AudioInL;
  sc_port<i_long_long_receiver> AudioInR;
  sc_port<i_long_long_sender>   VideoOut;
  sc_port<i_long_long_sender>   AudioOutLeft;
  sc_port<i_long_long_sender>   AudioOutRight;

  VideoCodec video;
  AudioCodec left, right;

  void before_end_of_elaboration()
  {
    video.p1.bind(VideoIn);
    video.p2.bind(VideoOut);

    left.p1.bind(AudioInL);
    left.p2.bind(AudioOutLeft);

    right.p1.bind(AudioInR);
    right.p2.bind(AudioOutRight);
  }

  SC_CTOR(DUT): video("video"), left("left"), right("right")
  { }
};

SC_MODULE(Display)
{
  sc_port<i_long_long_receiver> VideoIn;

  void display(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar('V');	// display the frame
    fflush(stdout);
#endif
  }

  void main(void)
  {
    long long Frame;

    while(1)
    {
      Frame = VideoIn->receive();
      display(Frame);
    }
  }

  SC_CTOR(Display)
  {
    SC_THREAD(main);
  }
};

SC_MODULE(Speaker)
{
  sc_port<i_long_long_receiver> AudioIn;

  char Channel;

  void sound(long long /* Frame */)
  {
#ifdef PRINT_FRAMES
    putchar(Channel); // output the sound
    fflush(stdout);
#endif
  }

  void main(void)
  {
    long long Frame;

    while(1)
    {
      Frame = AudioIn->receive();
      sound(Frame);
    }
  }

  Speaker(sc_module_name name, char chn): sc_module(name), Channel(chn)
  {
    SC_THREAD(main);
  }
  SC_HAS_PROCESS(Speaker);
};

SC_MODULE(Monitor)
{
  sc_port<i_long_long_receiver> Video;
  sc_port<i_long_long_receiver> AudioLeft;
  sc_port<i_long_long_receiver> AudioRight;

  Display screen;
  Speaker speakerL, speakerR;

  void before_end_of_elaboration()
  {
    screen.VideoIn.bind(Video);
    speakerL.AudioIn.bind(AudioLeft);
    speakerR.AudioIn.bind(AudioRight);
  }

  SC_CTOR(Monitor):
	screen("screen"),
	speakerL("speakerL", 'L'),
	speakerR("speakerR", 'R')
  { }
};

SC_MODULE(Top)
{
  c_long_long_queue VideoStream;
  c_long_long_queue AudioStreamL;
  c_long_long_queue AudioStreamR;
  c_long_long_queue Video;
  c_long_long_queue AudioLeft;
  c_long_long_queue AudioRight;

  Stimulus       stimulus;
  DUT            dut;
  Monitor        monitor;

  void before_end_of_elaboration()
  {
    stimulus.VideoStream.bind(VideoStream);
    stimulus.AudioStreamL.bind(AudioStreamL);
    stimulus.AudioStreamR.bind(AudioStreamR);

    dut.VideoIn.bind(VideoStream);
    dut.AudioInL.bind(AudioStreamL);
    dut.AudioInR.bind(AudioStreamR);
    dut.VideoOut.bind(Video);
    dut.AudioOutLeft.bind(AudioLeft);
    dut.AudioOutRight.bind(AudioRight);

    monitor.Video.bind(Video);
    monitor.AudioLeft.bind(AudioLeft);
    monitor.AudioRight.bind(AudioRight);
  }

  SC_CTOR(Top)
  : VideoStream("VideoStream", 10ul)
  , AudioStreamL("AudioStreamL", 10ul)
  , AudioStreamR("AudioStreamR", 10ul)
  , Video("Video", 10ul)
  , AudioLeft("AudioLeft", 10ul)
  , AudioRight("AudioRight", 10ul)
  , stimulus("stimulus")
  , dut("dut")
  , monitor("monitor")
  { }
};

int sc_main(int argc, char* argv[])
{
	Top top("top");
	printf("Playing...\n");
	sc_start();
	return 0;
}

// EOF
