#include "ooo_instrumentation.h"

#define SYSTEMC_LW_HOME_ENV "SYSTEMC_LW_HOME"
#define SYSTEMC_OOP_HOME_ENV "SYSTEMC_OOP_HOME"

#define HELP_PAGE                                                            \
"risc: Recoding Infrastructure for SystemC (RISC) Compiler, V0.2.1\n"        \
"      (c) 2015 CECS, University of California, Irvine\n"                    \
"\n"                                                                         \
"Usage: risc [<option>...] <design> [<option>...]\n"                         \
"\n"                                                                         \
"Options:        Description:                    Default:\n"                 \
"-h|--help       print this usage info, quit     none\n"                     \
"-v|--verbose    increment verbosity level       silent\n"                   \
"-w|--warnings   increment warning level         no warnings\n"              \
"-g|-g<level>    produce debugging information   no debugging infos\n"       \
"-O|-O<level>    enable optimizations            no optimization\n"          \
"-I<dir>         add dir to include path         $SYSTEMC_LW_HOME/include\n" \
"-L<dir>         add dir to library path         $SYSTEMC_OOP_HOME/lib\n"    \
"-l<lib>         add lib to linker libraries     -lsystemc\n"                \
"-c              compile only, do not link       compile and link\n"         \
"-o <filename>   name output file as filename    a.out\n"                    \
"<rose_option>   passed to ROSE compiler         none\n"                     \
"<g++_option>    passed to GNU C++ compiler      none\n"                     \
"\n"

namespace risc {

namespace inst {

InstrumentationTraversal::InstrumentationTraversal(SgType* variableType,
                                                   SgExpression* PrimChnlVarRefExp,
                                                   SgExpression* SCChnlVarRefExp,
                                                   SgClassDefinition* SCModuleClassDef,
                                                   std::map<SgNode*, int>& FuncCallExptoID)
  : variableType(variableType)
  , PrimChnlVarRefExp(PrimChnlVarRefExp)
  , SCChnlVarRefExp(SCChnlVarRefExp)
  , SCModuleClassDef(SCModuleClassDef)
  , FuncCallExptoID(FuncCallExptoID)
{}

void
InstrumentationTraversal::visit(SgNode* astNode)
{
  SgMemberFunctionDeclaration* memfuncdecl 
    = isSgMemberFunctionDeclaration(astNode);

  if (memfuncdecl != NULL)
  {
    // Instrument channel classes derived from ::sc_core::sc_prim_channel

    if (risc::tools::is_derived_from_sc_prim_channel(memfuncdecl->
    get_class_scope()))
    {

      // Due to ROSE limitation in template handling, our code instrumentation 
      // cannont process template member functions correctly. Refer to Section 
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl 
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s that is derived from ::sc_core::sc_prim_channel.\n",
                templmemfuncdecl->get_name().str(), 
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier 
          = memfuncdecl->get_specialFunctionModifier();

        if ( !memfuncmodifier.isConstructor() && // except constructors
             !memfuncmodifier.isDestructor() ) // except destructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgVariableDeclaration* variableDeclaration 
                = SageBuilder::buildVariableDeclaration ("_sc_chnl_scoped_lock", 
                  variableType, SageBuilder::buildCompoundInitializer(
                  SageBuilder::buildExprListExp(PrimChnlVarRefExp), 
                  PrimChnlVarRefExp->get_type()));

              SageInterface::prependStatement(variableDeclaration, block);
            }
          }
        }
      }
    }

    // Instrument channel classes derived from ::sc_core::sc_channel

    if (risc::tools::is_derived_from_sc_channel(memfuncdecl->
        get_class_scope()))
    {

      // Due to ROSE limitation in template handling, our code instrumentation 
      // cannont process template member functions correctly. Refer to Section 
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl 
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s that is derived from ::sc_core::sc_channel.\n", 
                templmemfuncdecl->get_name().str(), 
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf 
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier 
          = memfuncdecl->get_specialFunctionModifier();

        if ( !memfuncmodifier.isConstructor() && // except constructors
             !memfuncmodifier.isDestructor() ) // except destructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgVariableDeclaration* variableDeclaration 
                = SageBuilder::buildVariableDeclaration ("_sc_chnl_scoped_lock", 
                  variableType, SageBuilder::buildCompoundInitializer(
                  SageBuilder::buildExprListExp(SCChnlVarRefExp), 
                  SCChnlVarRefExp->get_type()));

              SageInterface::prependStatement(variableDeclaration, block);
            }
          }
        }
      }
    }

    // Instrument module classes derived from ::sc_core::sc_module

    if (risc::tools::is_derived_from_sc_module(memfuncdecl->get_class_scope()) || 
        risc::tools::is_derived_from_sc_channel(memfuncdecl->get_class_scope()))
    {

      // Due to ROSE limitation in template handling, our code instrumentation 
      // cannont process template member functions correctly. Refer to Section 
      // 5.6 on Page 64 in ROSE user manual (v. 0.9.6a, Feb 25, 2015)
      SgTemplateMemberFunctionDeclaration* templmemfuncdecl 
        = isSgTemplateMemberFunctionDeclaration(astNode);
      if (templmemfuncdecl != NULL)
      {
        fprintf(stderr, "Function %s is a template member function in Class %s that is derived from ::sc_core::sc_module.\n",
                templmemfuncdecl->get_name().str(), 
                templmemfuncdecl->get_class_scope()->get_declaration()->
                get_name().str());
        fprintf(stderr, "As ROSE/EDG allows no transformations on the template declaration, code instrumentation quits!\n");
        exit(1);
      }

      // Why not get function defining declaration? Then it will visit the same
      // function definition more than once
      SgFunctionDefinition* funcdf
        = isSgFunctionDefinition(memfuncdecl->get_definition());
      if (funcdf != NULL)
      {
        SgSpecialFunctionModifier memfuncmodifier 
          = memfuncdecl->get_specialFunctionModifier();

        if (memfuncmodifier.isConstructor()) // constructors
        {
          SgBasicBlock* block = funcdf->get_body();

          if (block != NULL)
          {
            if (block->get_statements().size() >= 1)
            {
              SgExprStatement* assignStmt = SageBuilder::buildAssignStatement(
                SageBuilder::buildVarRefExp("m_instance_id", SCModuleClassDef),
                SageBuilder::buildPlusPlusOp(SageBuilder::buildVarRefExp("_OoO_Instance_Counter", 
                memfuncdecl->get_class_scope()), SgUnaryOp::postfix));

              SageInterface::prependStatement(assignStmt, block);
            }
          }          
        }
      }
    }
  }

  // Instrument Instance Counters
  SgClassDefinition* classdf = isSgClassDefinition(astNode);

  if (classdf != NULL)
  {
    if (risc::tools::is_derived_from_sc_module(classdf) ||
        risc::tools::is_derived_from_sc_channel(classdf) )
    {
      // static unsigned int InstanceCounter
      SgVariableDeclaration* InstCntVarDecl
        = SageBuilder::buildVariableDeclaration ("_OoO_Instance_Counter", 
            SageBuilder::buildUnsignedIntType());
      InstCntVarDecl->get_declarationModifier().get_storageModifier().setStatic();

      SageInterface::prependStatement(InstCntVarDecl, classdf);

      // initialize InstanceCounter
      SgVariableDeclaration* InstCntInitDecl
        = SageBuilder::buildVariableDeclaration ("_OoO_Instance_Counter", 
            SageBuilder::buildUnsignedIntType(),
          SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(0)), classdf);

      SageInterface::insertStatementAfter(classdf->get_declaration(), InstCntInitDecl);
    }
  }

  // Instrument segment boundaries
  SgFunctionCallExp* funccallexp = isSgFunctionCallExp(astNode);
  if (funccallexp != NULL)
  {
    SgFunctionDeclaration* funcdecl = funccallexp->getAssociatedFunctionDeclaration();

    if (funcdecl != NULL)
    {
      if (funcdecl->get_name().getString() == "wait" ||
          funcdecl->get_name().getString() == "create_thread_process" ||
          funcdecl->get_name().getString() == "create_method_process" ||
          funcdecl->get_name().getString() == "create_cthread_process")
        if (FuncCallExptoID.find(funccallexp) != FuncCallExptoID.end())
          funccallexp->append_arg(SageBuilder::buildIntVal(FuncCallExptoID[funccallexp]));
        else
          funccallexp->append_arg(SageBuilder::buildIntVal(0));

      if (funcdecl->get_name().getString() == "create_thread_process" ||
          funcdecl->get_name().getString() == "create_method_process" ||
          funcdecl->get_name().getString() == "create_cthread_process")
        funccallexp->append_arg(SageBuilder::buildVarRefExp("m_instance_id", SCModuleClassDef));
    }
  }
}

int
risc_backend (SgProject* project, UnparseFormatHelp *unparseFormatHelp, UnparseDelegate* unparseDelagate,
      unsigned int VerbosityLevel) //....TODO risc::inst::ParseStatus status // hack by RD, 10/30/15, FIXME
{
  // DQ (7/12/2005): Introduce tracking of performance of ROSE.
  TimingPerformance timer ("AST Object Code Generation (backend):");

  int finalCombinedExitStatus = 0;

  if (project->get_binary_only() == true)
  {
    ROSE_ASSERT(project != NULL);

    // DQ (8/21/2008): Only output a message when we we use verbose option.
    if ( SgProject::get_verbose() >= 1 )
      printf ("Note: Binary executables are unparsed, but not passed to gcc as assembly source code \n");

    project->skipfinalCompileStep(true);
  }

  if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
    printf ("Inside of backend(SgProject*) \n");

  // printf ("   project->get_useBackendOnly() = %s \n",project->get_useBackendOnly() ? "true" : "false");
  if (project->get_useBackendOnly() == false)
  {
    // Add forward references for instantiated template functions and member functions 
    // (which are by default defined at the bottom of the file (but should be declared 
    // at the top once we know what instantiations should be built)).  They must be 
    // defined at the bottom since they could call other functions not yet declared in 
    // the file.  Note that this fixup is required since we have skipped the class template 
    // definitions which would contain the declarations that we are generating.  We might 
    // need that as a solution at some point if this fails to be sufficently robust.
    // if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL-2 )
    //      printf ("Calling fixupInstantiatedTemplates() \n");
    // DQ (9/6/2005): I think this is handled separately within post processing
    // (either that or they are just marked for output n the post processing)
    // fixupInstantiatedTemplates(project);

    // generate C++ source code
    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("Calling project->unparse() \n");

    project->unparse(unparseFormatHelp,unparseDelagate);

    /* Patch the static_cast errors (Guantao Liu, 07/29/2015) */
    char SysCmd[300];

//  std::cout << "\nStart patching the static_cast errors in the generated source files\n";
    if (VerbosityLevel >= 1 /* .... RISC_VERBOSE */)
    { std::cout << "RISC: patching ROSE limitations..."
            << std::endl;
    }
    for (int i = 0; i < project->numberOfFiles(); i++)
    {
      sprintf(SysCmd, "sed -r -i \'s/create_method_process\\s*\\((.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_method_process (\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, \\5, \\6, \\7)/\' rose_%s", (*project)[i]->get_sourceFileNameWithoutPath().c_str());

      if (VerbosityLevel >= 2 /* .... RISC_VERY_VERBOSE */)
      { std::cout << "  " << SysCmd << std::endl;
      }
      if (system(SysCmd) != 0)
      {
        std::cout << "\nERROR: Patch File ";
        std::cout << (*project)[i]->get_sourceFileNameWithoutPath();
        std::cout << "failed" << std::endl;
        exit(1);
      }

      sprintf(SysCmd, "sed -r -i \'s/create_thread_process\\s*\\((.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_thread_process (\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, \\5, \\6, \\7)/\' rose_%s", (*project)[i]->get_sourceFileNameWithoutPath().c_str());

      if (VerbosityLevel >= 2 /* .... RISC_VERY_VERBOSE */)
      { std::cout << "  " << SysCmd << std::endl;
      }
      if (system(SysCmd) != 0)
      {
        std::cout << "\nERROR: Patch File ";
        std::cout << (*project)[i]->get_sourceFileNameWithoutPath();
        std::cout << "failed" << std::endl;
        exit(1);
      }

      sprintf(SysCmd, "sed -r -i \'s/create_cthread_process\\s*\\((.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\,(.*)\\)/create_cthread_process (\\1, \\2, static_cast<sc_core::SC_ENTRY_FUNC>(\\3), \\4, \\5, \\6, \\7)/\' rose_%s", (*project)[i]->get_sourceFileNameWithoutPath().c_str());

      if (VerbosityLevel >= 2 /* .... RISC_VERY_VERBOSE */)
      { std::cout << "  " << SysCmd << std::endl;
      }
      if (system(SysCmd) != 0)
      {
        std::cout << "\nERROR: Patch File ";
        std::cout << (*project)[i]->get_sourceFileNameWithoutPath();
        std::cout << "failed" << std::endl;
        exit(1);
      }
    }

    /* End */

    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      std::cout << "  " << "source file(s) generated" << std::endl;
  }

  if (VerbosityLevel >= 1 /* .... RISC_VERBOSE */)
  { std::cout << "RISC: compiling executable..."
            << std::endl;
  }
  // DQ (1/25/2010): We have to now test for both numberOfFiles() and numberOfDirectories(),
  // or perhaps define a more simple function to use more directly.
  // if (project->numberOfFiles() > 0)
  if (project->numberOfFiles() > 0 || project->numberOfDirectories() > 0)
  {
    // Compile generated C++ source code with vendor compiler.
    // Generate object file (required for further template processing 
    // if templates exist).
    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("Calling project->compileOutput() \n");

      finalCombinedExitStatus = project->compileOutput();
  }
  else
  // if (project->get_compileOnly() == false)
  {
    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("   project->get_compileOnly() = %s \n",project->get_compileOnly() ? "true" : "false");

    // DQ (5/20/2005): If we have not permitted templates to be instantiated during initial 
    // compilation then we have to do the prelink step (this is however still new and somewhat 
    // problematic (buggy?)).  It relies upon the EDG mechansisms which are not well understood.
    bool callTemplateInstantation = (project->get_template_instantiation_mode() == SgProject::e_none);

    if (callTemplateInstantation == true)
    {
      // DQ (9/6/2005): I think that this is no longer needed
      printf ("I don't think we need to call instantiateTemplates() any more! \n");
      ROSE_ASSERT(false);

      // The instantiation of templates can cause new projects (sets of source files) 
      // to be generated, but since the object files are already processed this is 
      // not an issue here.  A project might, additionally, keep track of the ASTs
      // associated with the different phases of instantions of templates.
      if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
        printf ("Calling instantiateTemplates() \n");

       printf ("Skipping template support in backend(SgProject*) \n");
       // instantiateTemplates (project);
    }

    if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
      printf ("Calling project->link() \n");

    // DQ (10/15/2005): Trap out case of C programs where we want to make sure that we don't use the C++ compiler to do our linking!
    // This could be done in the 
    if (project->get_C_only() == true)
    {
      printf ("Link using the C language linker (when handling C programs) = %s \n","gcc");
      // finalCombinedExitStatus = project->link("gcc");
      finalCombinedExitStatus = project->link("gcc");
    }
    else
    {
      // Use the default name for C++ compiler (defined at configure time)
      if ( SgProject::get_verbose() >= BACKEND_VERBOSE_LEVEL )
        printf ("Link using the default linker (when handling non-C programs) = %s \n","g++");
      finalCombinedExitStatus = project->link("g++");
    }

    // printf ("DONE with link! \n");
  }

  // Message from backend to user.
  // Avoid all I/O to stdout if useBackendOnly == true.
  // if (project->get_useBackendOnly() == false)
  if ( SgProject::get_verbose() >= 1 )
    std::cout << "source file(s) compiled with vendor compiler. (exit status = " << finalCombinedExitStatus << ").\n" << std::endl;

  // Set the final error code to be returned to the user.
  project->set_backendErrorCode(finalCombinedExitStatus);
  return project->get_backendErrorCode();
}

int
instrumentor (risc::Design& design, risc::sg::DataConflictTable* data_conflict_table, 
  risc::sg::EventConflictTable* event_conflict_table,
  risc::sg::TimeAdvancedTable* time_advance_table,
  std::map<SgNode*, int>& FuncCallExptoID,
      unsigned int VerbosityLevel) //....TODO risc::inst::ParseStatus status // hack by RD, 10/30/15, FIXME
{
  SgProject* project = design.project_;
  ROSE_ASSERT (project != NULL);
  SgClassDeclaration* ChnlScpLockClassDecl = NULL;
  SgFunctionDeclaration* SCMainFuncDecl = NULL;
  SgType* variableType = NULL;
  SgExpression* PrimChnlVarRefExp = NULL;
  SgExpression* SCChnlVarRefExp = NULL;
  SgClassDefinition* PrimChnlClassDef = NULL;
  SgClassDefinition* SCChnlClassDef = NULL;
  SgClassDefinition* SCModuleClassDef = NULL;
  std::string LWIncludePath;
  std::string LWLibraryPath;
  std::string OOPIncludePath;
  std::string OOPLibraryPath;
  std::string OOPLibraryPathwoL;


  // Find the class declarations of chnl_scoped_lock, sc_prim_channel, 
  // sc_channel and sc_module
  Rose_STL_Container<SgNode*> classDefinitionList 
    = NodeQuery::querySubTree(project, V_SgClassDefinition);
  for (Rose_STL_Container<SgNode*>::iterator it = classDefinitionList.begin(); 
       it != classDefinitionList.end(); it++)
  {
    SgClassDefinition* classdf = isSgClassDefinition(*it);
    if (classdf != NULL)
    {
      if (classdf->get_qualified_name().getString() == 
          "::sc_core::chnl_scoped_lock")
      {
        // SgClassDeclaration of "::sc_core::chnl_scoped_lock"
        ChnlScpLockClassDecl = classdf->get_declaration();  
      }

      if (classdf->get_qualified_name().getString() == 
          "::sc_core::sc_prim_channel")
      {
        // SgClassDefinition of "::sc_core::sc_prim_channel"
        PrimChnlClassDef = classdf;  
      }

      if (classdf->get_qualified_name().getString() == 
          "::sc_core::sc_channel")
      {
        // SgClassDefinition of "::sc_core::sc_channel"
        SCChnlClassDef = classdf;  
      }

      if (classdf->get_qualified_name().getString() == 
          "::sc_core::sc_module")
      {
        // SgClassDefinition of "::sc_core::sc_module"
        SCModuleClassDef = classdf;  
      }

      if (ChnlScpLockClassDecl && PrimChnlClassDef &&
          SCChnlClassDef && SCModuleClassDef)
        break;
    }
  }
  ROSE_ASSERT(ChnlScpLockClassDecl != NULL);
  ROSE_ASSERT(PrimChnlClassDef != NULL);
  ROSE_ASSERT(SCChnlClassDef != NULL);
  ROSE_ASSERT(SCModuleClassDef != NULL);

  variableType = ChnlScpLockClassDecl->get_type();
  ROSE_ASSERT(variableType != NULL);

  PrimChnlVarRefExp = SageBuilder::buildVarRefExp("m_mutex", PrimChnlClassDef);
  ROSE_ASSERT(PrimChnlVarRefExp != NULL);

  SCChnlVarRefExp = SageBuilder::buildVarRefExp("m_mutex", SCChnlClassDef);
  ROSE_ASSERT(SCChnlVarRefExp != NULL);

  // Find the function declaration of sc_main
  Rose_STL_Container<SgNode*> funcDeclarationList 
    = NodeQuery::querySubTree(project, V_SgFunctionDeclaration);
  for (Rose_STL_Container<SgNode*>::iterator it = funcDeclarationList.begin(); 
       it != funcDeclarationList.end(); it++)
  {
    SgFunctionDeclaration* funcdecl = isSgFunctionDeclaration(*it);
    if (funcdecl != NULL)
    {
      if (funcdecl->get_name().getString() == "sc_main"
        && funcdecl->get_definition() != NULL)
      {
        // SgFunctionDeclaration of "sc_main"
        SCMainFuncDecl = funcdecl;  
        break;
      }
    }
  }
  ROSE_ASSERT(SCMainFuncDecl != NULL);

  // Build the traversal object
  InstrumentationTraversal traversal(variableType, PrimChnlVarRefExp, 
                                     SCChnlVarRefExp, SCModuleClassDef,
                                     FuncCallExptoID);

  // Call the traversal starting at the project node of the AST
  traversal.traverseInputFiles(project, preorder);

  // Output the data conflict table
  int DataConflTableSize = data_conflict_table->size_of_conflict_table_;
  std::vector<SgExpression *> DataConflTableValue;
  for (int i = 0; i < DataConflTableSize; i++)
    for (int j = 0; j < DataConflTableSize; j++)
      DataConflTableValue.push_back(SageBuilder::buildIntVal(data_conflict_table->conflict_table_[i][j]));

  SgVariableDeclaration* DataConflTableDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Data_Conflict_Table",
      SageBuilder::buildArrayType(SageBuilder::buildBoolType(),
      SageBuilder::buildIntVal(DataConflTableSize * DataConflTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(DataConflTableValue)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, DataConflTableDecl);

  // Output the data conflict table size
  SgVariableDeclaration* DataConflTableSizeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Data_Conflict_Table_Size",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(DataConflTableSize)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, DataConflTableSizeDecl);

  // Output the event notification table
  int EventNotifyTableSize = event_conflict_table->size_of_conflict_table_;
  std::vector<SgExpression *> EventNotifyTableValue;
  for (int i = 0; i < EventNotifyTableSize; i++)
    for (int j = 0; j < EventNotifyTableSize; j++)
      EventNotifyTableValue.push_back(SageBuilder::buildIntVal(event_conflict_table->conflict_table_[i][j]));

  SgVariableDeclaration* EventNotifyTableDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Event_Notify_Table",
      SageBuilder::buildArrayType(SageBuilder::buildBoolType(), 
      SageBuilder::buildIntVal(EventNotifyTableSize * EventNotifyTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(EventNotifyTableValue)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, EventNotifyTableDecl);

  // Output the event notification table size
  SgVariableDeclaration* EventNotifyTableSizeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Event_Notify_Table_Size",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(EventNotifyTableSize)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, EventNotifyTableSizeDecl);

  // Output the segment and instance id to index lookup table for conflict tables
  int MaxNumofInst = data_conflict_table->max_instances_;
  int NumofSeg = data_conflict_table->number_of_segments_;
  std::vector<SgExpression *> ConflictLookupTableValue;
  for (int i = 0; i < MaxNumofInst + 1; i++)
    for (int j = 0; j < NumofSeg; j++)
    {
      //ConflictLookupTableValue.push_back(buildIntVal(
        //data_conflict_table->instance_id_to_index_lookup_[i][j]));

      // Temporal patch: all unanalyzed segment boundaries are given ID 0 TODO
      if (j == 0 && data_conflict_table->instance_id_to_index_lookup_[i][j] == -1)
        ConflictLookupTableValue.push_back(SageBuilder::buildIntVal(0));
      else
        ConflictLookupTableValue.push_back(SageBuilder::buildIntVal(
          data_conflict_table->instance_id_to_index_lookup_[i][j]));
    }

  SgVariableDeclaration* ConflictLookupTableDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Conflict_Index_Lookup_Table",
      SageBuilder::buildArrayType(SageBuilder::buildIntType(),
      SageBuilder::buildIntVal((MaxNumofInst + 1) * NumofSeg)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(ConflictLookupTableValue)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, ConflictLookupTableDecl);

  // Output the maximum number of module instances and number of segments
  SgVariableDeclaration* MaxNumofInstDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Max_Number_of_Instances",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(MaxNumofInst)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, MaxNumofInstDecl);

  SgVariableDeclaration* NumofSegDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Number_of_Segments",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(NumofSeg)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, NumofSegDecl);

  // Output the current time advance table
  int CurrTimeAdvTableSize = time_advance_table->size_;

  std::vector<SgExpression *> CurrTimeAdvTableTimeValue;
  for (int i = 0; i < CurrTimeAdvTableSize; i++)
  {
    if (time_advance_table->current_time_advance_[i].infinite_)
      CurrTimeAdvTableTimeValue.push_back(SageBuilder::buildLongLongIntVal(-1)); // -1 stands for infinity
    else
      CurrTimeAdvTableTimeValue.push_back(SageBuilder::buildLongLongIntVal(
        time_advance_table->current_time_advance_[i].time_units_));
  }

  SgVariableDeclaration* CurrTimeAdvTableTimeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Curr_Time_Advance_Table_Time",
      SageBuilder::buildArrayType(SageBuilder::buildLongLongType(), 
                                  SageBuilder::buildIntVal(CurrTimeAdvTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(CurrTimeAdvTableTimeValue)));

  std::vector<SgExpression *> CurrTimeAdvTableDeltaValue;
  for (int i = 0; i < CurrTimeAdvTableSize; i++)
  {
    if (time_advance_table->current_time_advance_[i].infinite_)
      CurrTimeAdvTableDeltaValue.push_back(SageBuilder::buildIntVal(-1)); // -1 stands for infinity
    else
      CurrTimeAdvTableDeltaValue.push_back(SageBuilder::buildIntVal(
        time_advance_table->current_time_advance_[i].delta_count_));
  }

  SgVariableDeclaration* CurrTimeAdvTableDeltaDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Curr_Time_Advance_Table_Delta",
      SageBuilder::buildArrayType(SageBuilder::buildIntType(), 
                                  SageBuilder::buildIntVal(CurrTimeAdvTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(CurrTimeAdvTableDeltaValue)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, CurrTimeAdvTableTimeDecl);
  SageInterface::insertStatementBefore(SCMainFuncDecl, CurrTimeAdvTableDeltaDecl);

  // Output the current time advance table size
  SgVariableDeclaration* CurrTimeAdvTableSizeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Curr_Time_Advance_Table_Size",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(CurrTimeAdvTableSize)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, CurrTimeAdvTableSizeDecl);

  // Output the next time advance table
  int NextTimeAdvTableSize = time_advance_table->size_;

  std::vector<SgExpression *> NextTimeAdvTableTimeValue;
  for (int i = 0; i < NextTimeAdvTableSize; i++)
  {
    if (time_advance_table->next_time_advance_[i].infinite_)
      NextTimeAdvTableTimeValue.push_back(SageBuilder::buildLongLongIntVal(-1)); // -1 stands for infinity
    else
      NextTimeAdvTableTimeValue.push_back(SageBuilder::buildLongLongIntVal(
        time_advance_table->next_time_advance_[i].time_units_));
  }

  SgVariableDeclaration* NextTimeAdvTableTimeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Next_Time_Advance_Table_Time",
      SageBuilder::buildArrayType(SageBuilder::buildLongLongType(), 
                                  SageBuilder::buildIntVal(NextTimeAdvTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(NextTimeAdvTableTimeValue)));

  std::vector<SgExpression *> NextTimeAdvTableDeltaValue;
  for (int i = 0; i < NextTimeAdvTableSize; i++)
  {
    if (time_advance_table->next_time_advance_[i].infinite_)
      NextTimeAdvTableDeltaValue.push_back(SageBuilder::buildIntVal(-1)); // -1 stands for infinity
    else
      NextTimeAdvTableDeltaValue.push_back(SageBuilder::buildIntVal(
        time_advance_table->next_time_advance_[i].delta_count_));
  }

  SgVariableDeclaration* NextTimeAdvTableDeltaDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Next_Time_Advance_Table_Delta",
      SageBuilder::buildArrayType(SageBuilder::buildIntType(), 
                                  SageBuilder::buildIntVal(NextTimeAdvTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(NextTimeAdvTableDeltaValue)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, NextTimeAdvTableTimeDecl);
  SageInterface::insertStatementBefore(SCMainFuncDecl, NextTimeAdvTableDeltaDecl);

  // Output the next time advance table size
  SgVariableDeclaration* NextTimeAdvTableSizeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Next_Time_Advance_Table_Size",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(NextTimeAdvTableSize)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, NextTimeAdvTableSizeDecl);

  // Output the segment id to index lookup table for time advance tables
  int TimeAdvLookupTableSize = time_advance_table->size_;
  std::vector<SgExpression *> TimeAdvLookupTableValue;
  for (int i = 0; i < TimeAdvLookupTableSize; i++)
      TimeAdvLookupTableValue.push_back(SageBuilder::buildIntVal(time_advance_table->get_segment_id(i)));

  SgVariableDeclaration* TimeAdvLookupTableDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Time_Advance_Index_Lookup_Table",
      SageBuilder::buildArrayType(SageBuilder::buildIntType(), 
                                  SageBuilder::buildIntVal(TimeAdvLookupTableSize)), 
      SageBuilder::buildAggregateInitializer(SageBuilder::buildExprListExp(TimeAdvLookupTableValue)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, TimeAdvLookupTableDecl);

  // Output the size of the segment id to index lookup table
  SgVariableDeclaration* TimeAdvLookupTableSizeDecl 
    = SageBuilder::buildVariableDeclaration ("sc_core::_OoO_Time_Advance_Index_Lookup_Table_Size",
      SageBuilder::buildUnsignedIntType(), 
      SageBuilder::buildAssignInitializer(SageBuilder::buildIntVal(TimeAdvLookupTableSize)));

  SageInterface::insertStatementBefore(SCMainFuncDecl, TimeAdvLookupTableSizeDecl);

  // Test the new AST
  AstTests::runAllTests(project);

  // Change the command line argument list
  LWIncludePath = "-I";
  LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
  LWIncludePath += "/include";
  LWLibraryPath = "-L";
  LWLibraryPath += getenv(SYSTEMC_LW_HOME_ENV);
  LWLibraryPath += "/lib-linux64";
  OOPIncludePath = "-I";
  OOPIncludePath += getenv(SYSTEMC_OOP_HOME_ENV);
  OOPIncludePath += "/include";
  OOPLibraryPath = "-L";
  OOPLibraryPath += getenv(SYSTEMC_OOP_HOME_ENV);
  OOPLibraryPath += "/lib-linux64";
  OOPLibraryPathwoL = OOPLibraryPath;
  OOPLibraryPathwoL.erase(0, 2);

  Rose_STL_Container<std::string> argList = project->get_originalCommandLineArgumentList();
  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    if (it->compare(LWIncludePath) == 0)
      *it = OOPIncludePath;

    if (it->compare(LWLibraryPath) == 0)
      *it = OOPLibraryPath;
  }
  argList.push_back(std::string("-Xlinker"));
  argList.push_back(std::string("-R"));
  argList.push_back(std::string("-Xlinker"));
  argList.push_back(OOPLibraryPathwoL);
  project->set_originalCommandLineArgumentList(argList);

  for (int i = 0; i < project->numberOfFiles(); i++)
  {
    SgFile& file = project->get_file(i);
    Rose_STL_Container<std::string> argListFile = file.get_originalCommandLineArgumentList();
    for (Rose_STL_Container<std::string>::iterator it = argListFile.begin();
         it != argListFile.end(); it++)
    {
      if (it->compare(LWIncludePath) == 0)
        *it = OOPIncludePath;

      if (it->compare(LWLibraryPath) == 0)
        *it = OOPLibraryPath;
    }
    argListFile.push_back(std::string("-Xlinker"));
    argListFile.push_back(std::string("-R"));
    argListFile.push_back(std::string("-Xlinker"));
    argListFile.push_back(OOPLibraryPathwoL);
    file.set_originalCommandLineArgumentList(argListFile);
  }

  return risc_backend(project, NULL, NULL, VerbosityLevel); // .... FIXME hack by RD
}

ParseStatus::ParseStatus()
{
  VerbosityLevel = 0;
  WarningLevel = 0;
}

ParseStatus& ParseStatus::operator=(const ParseStatus& copy)
{
  VerbosityLevel = copy.VerbosityLevel;
  WarningLevel = copy.WarningLevel;

  return *this;
}

ParseStatus
ParseCommandlineOptions(int &argc, char** &argv)
{
  std::string LWIncludePath;
  std::string LWLibraryPath;

  LWIncludePath = "-I";
  LWIncludePath += getenv(SYSTEMC_LW_HOME_ENV);
  LWIncludePath += "/include";
  LWLibraryPath = "-L";
  LWLibraryPath += getenv(SYSTEMC_LW_HOME_ENV);
  LWLibraryPath += "/lib-linux64";

  ParseStatus status;

  if (argc < 2)
  {
    std::cout << "risc: no input files\n";
    exit(0);
  }

  Rose_STL_Container<std::string> argList;
  argList = CommandlineProcessing::generateArgListFromArgcArgv(argc, argv);

  std::vector<Rose_STL_Container<std::string>::iterator> RemovedIter;

  for (Rose_STL_Container<std::string>::iterator it = argList.begin();
       it != argList.end(); it++)
  {
    size_t pos;

    // -h, --help
    if (it->compare("-h") == 0 || it->compare("--help") == 0)
    {
      std::cout << HELP_PAGE;
      exit(0);
    }

    // -v, --verbose
    if (it->compare("-v") == 0 || it->compare("--verbose") == 0)
    {
      status.VerbosityLevel ++;
      RemovedIter.push_back(it);
    }

    // -vv
    if (it->compare("-vv") == 0)
    {
      status.VerbosityLevel += 2;
      RemovedIter.push_back(it);
    }

    // -vvv
    if (it->compare("-vvv") == 0)
    {
      status.VerbosityLevel += 3;
      RemovedIter.push_back(it);
    }

    // -w, --warnings
    if (it->compare("-w") == 0 || it->compare("--warnings") == 0)
    {
      status.WarningLevel ++;
      RemovedIter.push_back(it);
    }

    // -ww
    if (it->compare("-ww") == 0)
    {
      status.WarningLevel += 2;
      RemovedIter.push_back(it);
    }

    // -www
    if (it->compare("-www") == 0)
    {
      status.WarningLevel += 3;
      RemovedIter.push_back(it);
    }
  }

  while (RemovedIter.size() != 0)
  {
    argList.erase(RemovedIter.back());
    RemovedIter.pop_back();
  }

  // add -I$SYSTEMC_LW_HOME/include
  argList.push_back(LWIncludePath);

  // add -L$SYSTEMC_LW_HOME/lib-linux64
  argList.push_back(LWLibraryPath);

  // add -lsystemc
  argList.push_back(std::string("-lsystemc"));

  argc = 0;
  argv = NULL;

  CommandlineProcessing::generateArgcArgvFromList(argList, argc, argv);

  return status;
}

} // end of namespace inst

} // end of namespace risc

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
