#include "conflict_table.h"

#include<vector>
#include<utility>
#include<iomanip>

#include "graph_helper.h"

risc::sg::ConflictTable::ConflictTable(
    SegmentGraph &segment_graph,
    PathInstanceMapper &path_instance_mapper,
    bool debugging):
  segment_graph_(segment_graph),
  path_instance_mapper_(path_instance_mapper),
  debugging_(debugging)
{
  number_of_segments_ = boost::num_vertices(segment_graph_.graph_);
  max_instances_ = path_instance_mapper_.max_instances_;

  instance_id_to_index_lookup_ = new int*[max_instances_ + 1];

  // We are adding one additional row to safe the segment id
  for(int i=0; i < max_instances_ + 1; i++) {

    instance_id_to_index_lookup_[i]
      = new int[number_of_segments_];
  }

  // Here, the first row of the instance_id_to_index_lookup table is filled
  // with the segment ids. We have to do that because the segment ids do not
  // incease by one.
  std::pair<VertexIterator, VertexIterator> vp;
  int i = 0;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    instance_id_to_index_lookup_[0][i] = segment_graph_.graph_[*(vp.first)].id_;
  }

  // We are filling all array fields with -1 to indicate that a specific field
  // is not used. We need that information for the regression tests.
  for(int i = 1; i < max_instances_ + 1; i++) {
    for(int j=0; j < number_of_segments_; j++) {
      instance_id_to_index_lookup_[i][j] = -1;
    }
  }

  // Here, we are filling the instance_id_to_index_lookup_ table. Each field
  // in this table needs an unique value. The value is the associated row/column
  // in the conflict table.
  int index = 0;

  // We are iterating over all the instances in the design
  for(std::vector<PathInstanceMapper::PathWithID>::iterator
      iter  = path_instance_mapper_.path_to_id_mapping_.begin();
      iter != path_instance_mapper_.path_to_id_mapping_.end();
      iter++) {

    // Adding the value of 1 because the first row is reserved for the
    // segment id
    int row = (*iter).second + 1;

    Object *object = (*iter).first.get_object();

    ModuleInstance *module_instance = dynamic_cast<ModuleInstance*>(object);

    if(module_instance) {

      Module *module = module_instance->definition_;

      for(ThreadVectorIter
          thread_iter  = module->threads_.begin();
          thread_iter != module->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorIter
          cthread_iter  = module->cthreads_.begin();
          cthread_iter != module->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorIter
          method_iter  = module->methods_.begin();
          method_iter != module->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }

    // This is the same like for the modules. So, please see the documents above
    HierarchicalChannelInstance *hierarchical_channel_instance
      = dynamic_cast<HierarchicalChannelInstance*>(object);

    if(hierarchical_channel_instance) {

      HierarchicalChannel *hierarchical_channel
        = hierarchical_channel_instance->definition_;

      for(ThreadVectorIter
          thread_iter  = hierarchical_channel->threads_.begin();
          thread_iter != hierarchical_channel->threads_.end();
          thread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*thread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(CThreadVectorIter
          cthread_iter  = hierarchical_channel->cthreads_.begin();
          cthread_iter != hierarchical_channel->cthreads_.end();
          cthread_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*cthread_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }

      for(MethodVectorIter
          method_iter  = hierarchical_channel->methods_.begin();
          method_iter != hierarchical_channel->methods_.end();
          method_iter++) {

        std::list<int> reachable_ids = get_all_reachable_segments(
            segment_graph, (*method_iter)->starting_id_);

        for(std::list<int>::iterator
            id_iter  = reachable_ids.begin();
            id_iter != reachable_ids.end();
            id_iter++) {

          for(int i=0; i < number_of_segments_; i++) {

            if(instance_id_to_index_lookup_[0][i] == *id_iter){

              if(instance_id_to_index_lookup_[row][i] == -1) {
                instance_id_to_index_lookup_[row][i] = index;
                index++;
              }

              break;
            }
          }
        }
      }
    }
  }

  size_of_conflict_table_ = index;


  // Allocating the memory for the VertexDescriptor array
  vertex_lookup_ = new VertexDescriptor[number_of_segments_];

  // Here, we are filling the vertex lookup. The array field at position is the
  // segment with the index instance_id_to_index_lookup_[0][i]
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first, i++) {

    int i = 0;
    for(; i < number_of_segments_; i++) {
      if(instance_id_to_index_lookup_[0][i]
          == segment_graph_.graph_[*(vp.first)].id_) {

        vertex_lookup_[i] = *(vp.first);
        break;
      }
    }
  }

  // creating the conflict table
  conflict_table_ = new bool*[size_of_conflict_table_];
  for(int i=0; i<size_of_conflict_table_; i++) {
    conflict_table_[i] = new bool[size_of_conflict_table_];
  }

  if(debugging_) {

    conflict_info_ = new ConflictInformation*[size_of_conflict_table_];
    for(int i=0; i<size_of_conflict_table_; i++) {
      conflict_info_[i] = new ConflictInformation[size_of_conflict_table_];
    }
  } else {

    conflict_info_ = NULL;
  }
}

void
risc::sg::ConflictTable::determine_conflict_table()
{
  for(int i=0; i<size_of_conflict_table_; i++) {
    for(int j=0; j<size_of_conflict_table_; j++) {

      conflict_table_[i][j] = false;
    }
  }
}

risc::sg::ConflictTable::~ConflictTable()
{
  for(int i=0; i < max_instances_ + 1; i++) {

    delete[] instance_id_to_index_lookup_[i];
  }
  delete[] instance_id_to_index_lookup_;


  for(int i=0; i < size_of_conflict_table_; i++) {

    delete[] conflict_table_[i];
  }

  delete[] conflict_table_;

  delete[] vertex_lookup_;

  if(debugging_) {
    for(int i=0; i < size_of_conflict_table_; i++) {

      delete[] conflict_info_[i];
    }
    delete[] conflict_info_;
  }
}

void
risc::sg::ConflictTable::print_instance_id_to_index_lookup_table()
{

  for(int i=0; i < max_instances_ + 1; i++) {

    if(i == 0) {

      std::cout << "    ";
    } else {
      std::cout << std::setw(4) << i-1;
    }

    for(int j=0; j < boost::num_vertices(segment_graph_.graph_); j++) {

      std::cout << std::setw(4) << instance_id_to_index_lookup_[i][j] << " ";
    }

    std::cout << std::endl;
  }
}

int
risc::sg::ConflictTable::segment_and_instance_id_to_index(
    int segment_id, int instance_id)
{
  int column_index = -1;
  for(int i = 0; i < number_of_segments_; i++) {

    if(instance_id_to_index_lookup_[0][i] == segment_id) {

      column_index = i;
    }
  }

  if(column_index == -1) {
    assert("segment_and_instance_id_to_index: This segment id does not exist"
        && 0);
  }

  return instance_id_to_index_lookup_ [instance_id + 1][column_index];
}

std::pair<int, int>
risc::sg::ConflictTable::index_to_segment_and_instance_id(int index)
{
  std::pair<int, int> result;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      if(instance_id_to_index_lookup_[i][j] == index) {

        result.first = instance_id_to_index_lookup_[0][j];
        result.second = i - 1;
        return result;
      }
    }
  }

  std::cout << "index: " << index << std::endl;
  assert("index_to_segment_and_instance_id: Index does not exist" && 0);
}

void
risc::sg::ConflictTable::set_conflict_cause(
    int segment_id_1, int instance_id_1, int segment_id_2, int instance_id_2,
    SgNode *cause)
{
  conflict_table_
    [segment_and_instance_id_to_index(segment_id_1, instance_id_1)]
    [segment_and_instance_id_to_index(segment_id_2, instance_id_2)] = true;

  if(debugging_) {

    if(cause) {

      conflict_info_
        [segment_and_instance_id_to_index(segment_id_1, instance_id_1)]
        [segment_and_instance_id_to_index(segment_id_2, instance_id_2)]
          .conflict_variables_.insert(cause);
    }
  }
}

bool
risc::sg::ConflictTable::has_conflict(int segment_id_1, int instance_id_1,
    int segment_id_2, int instance_id_2)
{
  return conflict_table_
    [segment_and_instance_id_to_index(segment_id_1, instance_id_1)]
    [segment_and_instance_id_to_index(segment_id_2, instance_id_2)];
}

void
risc::sg::ConflictTable::print_conflict_table(std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i = 1; i < max_instances_ + 1; i++) {

    if(i == 1)
      myfile << "    <th>(Seg,Inst)</th>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {


      if(instance_id_to_index_lookup_[i][j] == -1) {
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <th>"
             << "(" << ids.first << "," << ids.second << ")" << "</th>"
             << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i = 1; i < max_instances_ + 1; i++) {

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "  <tr>" << std::endl;

      if(instance_id_to_index_lookup_[i][j] == -1) {
        myfile << "  </tr>" << std::endl;
        continue;
      }

      std::pair<int,int> ids = index_to_segment_and_instance_id(
                                   instance_id_to_index_lookup_[i][j]);

      myfile << "    <td>"
             << "(" << ids.first << "," << ids.second << ")" << "</td>"
             << std::endl;

      for(int k = 1; k < max_instances_ + 1; k++) {

        for(int l=0; l < number_of_segments_; l++) {

          if(instance_id_to_index_lookup_[k][l] == -1) {
            continue;
          }

          int row = instance_id_to_index_lookup_[i][j];
          int column = instance_id_to_index_lookup_[k][l];

          if(conflict_table_[row] [column]) {

            myfile << "    <td class=\"tg-5fb6\">" << std::endl;

            if(debugging_) {
              for(std::set<SgNode*>::iterator
                  iter  = conflict_info_[row][column].
                              conflict_variables_.begin();
                  iter != conflict_info_[row][column].
                              conflict_variables_.end();
                  iter++) {

                SgVariableSymbol *var_sym = isSgVariableSymbol(*iter);
                if(var_sym) {

                  myfile << "      <pre>"
                         << var_sym->get_declaration()->unparseToString()
                         << "</pre>"  << std::endl;
                } else {
                  myfile << "      <pre>"
                         << (*iter)->unparseToString()
                         << "</pre>"  << std::endl;
                }
              }
            }

            myfile << "    </td>" << std::endl;
          } else {
            myfile << "    <td class=\"tg-thls\"></td>" << std::endl;
          }
        }
      }

      myfile << "  </tr>" << std::endl;
    }
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

void
risc::sg::ConflictTable::print_lookup_table(std::string filename)
{
  std::ofstream myfile;
  myfile.open(filename.c_str());
  myfile << "<!DOCTYPE html><html><head></head><body>" << std::endl;

  myfile << "<style type=\"text/css\">" << std::endl;
  myfile << ".tg  {border-collapse:collapse;border-spacing:0;}" << std::endl;
  myfile << ".tg td{font-family:Arial, sans-serif;font-size:14px;padding:"
         << "10px 10px;border-style:solid;border-width:1px;overflow:hidden;"
         << "word-break:normal;}"
         << std::endl;

  myfile << ".tg th{font-family:Arial, sans-serif;font-size:14px;font-weight:"
         << "normal;padding:10px 10px;border-style:solid;border-width:1px;"
         << "overflow:hidden;word-break:normal;}" << std::endl;
  myfile << ".tg .tg-thls{background-color:#009901}" << std::endl;
  myfile << ".tg .tg-5fb6{background-color:#fe0000}" << std::endl;
  myfile << "</style>" << std::endl;

  myfile << "<table class=\"tg\">" << std::endl;


  myfile << "  <tr>" << std::endl;
  for(int i=0; i < number_of_segments_ + 1; i++) {

    if(i == 0) {
      myfile << "    <th> </th>" << std::endl;

    } else {
      myfile << "    <th>"
             << instance_id_to_index_lookup_[0][i-1]
             << "</th>" << std::endl;
    }
  }
  myfile << "  </tr>" << std::endl;

  for(int i=1; i < max_instances_ + 1; i++) {

    myfile << "  <tr>" << std::endl;
    myfile << "<td>" << i-1 << "</td>" << std::endl;

    for(int j=0; j < number_of_segments_; j++) {

      myfile << "<td>" << instance_id_to_index_lookup_[i][j] << "</td>" << std::endl;
    }
    myfile << "  </tr>" << std::endl;
  }

  myfile << "</table>" << std::endl;

  myfile << "</body></html>";
  myfile.close();
}

risc::sg::VertexDescriptor
risc::sg::ConflictTable::segment_id_to_vertex_descriptor(int id)
{
  for(int i = 0; i < number_of_segments_; i++) {
    if(instance_id_to_index_lookup_[0][i] == id) {
      return vertex_lookup_[i];
    }
  }

  assert("segment_id_to_vertex_descriptor: Id doesn't exist" && 0);
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
