#include "data_conflict_table.h"

#include <fstream>
#include "../tools/variable_helper.h"
#include "../internal_representation/design.h"


risc::sg::DataConflictTable::DataConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper &path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  // here we perform a read/write analysis on each segment in the graph
  std::pair<VertexIterator, VertexIterator> vp;
  for(vp = boost::vertices(segment_graph_.graph_);
      vp.first != vp.second; ++vp.first) {

    segment_graph_.graph_[*vp.first].analyze_segment();
  }

  determine_conflict_table();
}

SgVariableSymbol*
risc::sg::DataConflictTable::get_mapped_variable(
    SgVariableSymbol *variable,
    int instance_id)
{
  if(isSgReferenceType(variable->get_type())) {

    SgVariableDefinition *var_def
      = isSgVariableDefinition(variable->get_declaration()
          ->get_definition());

    std::list<InstanceTree> potential_instances
      = risc::get_design()->get_root().find_all_instances_of(var_def);

    InstanceTree waiting_event_tree
      = path_instance_mapper_.get_instance_tree(
          instance_id, potential_instances);

    std::list<risc::InstanceTree> temp
      = waiting_event_tree.find_all_instances_of(var_def);

    return (temp.front().get_real_mapped_variable().get_object_symbol());

  } else {

    return variable;
  }
}

void
risc::sg::DataConflictTable::determine_conflict_table()
{
  for(int row=0; row < size_of_conflict_table_; row++) {

    // Here, we are getting the segment id and instance id with the associated
    // row index 'row'.
    std::pair<int, int> row_id_and_instance =
      index_to_segment_and_instance_id(row);

    int row_seg = row_id_and_instance.first;
    int row_inst = row_id_and_instance.second;

    for(int column=0; column <= row; column++) {

      // Here, we are getting the segment id and instance id with the associated
      // column index 'column'.
      std::pair<int, int> column_id_and_instance =
        index_to_segment_and_instance_id(column);

      int column_seg = column_id_and_instance.first;
      int column_inst = column_id_and_instance.second;

      if(column_seg == row_seg && row_inst == column_inst) {

        // If only read access in this segment, there will be no conflict
        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(column_seg)].
            write_variables_.empty()) {

          conflict_table_[row][column] = false;
        } else {

          bool has_conflict = false;
          for(std::set<SgVariableSymbol*>::iterator
              write_iter  = segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.begin();
              write_iter != segment_graph_.
                              graph_[segment_id_to_vertex_descriptor(row_seg)].
                                write_variables_.end();
              write_iter++) {

            if((tools::is_member_variable(*write_iter)) ||

                (tools::is_local_variable(*write_iter) &&
                tools::is_static_variable(*write_iter)) ||

                tools::is_global_variable(*write_iter) ||
                tools::is_namespace_variable(*write_iter)) {
              has_conflict = true;

              if(debugging_) {
                conflict_info_[column][row].conflict_variables_.
                  insert(*write_iter);
                conflict_info_[row][column].conflict_variables_.
                  insert(*write_iter);
              }
            }
          }

          conflict_table_[row][column] = has_conflict;
        }

        if(segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
            in_conflict_with_all_other_segments){

          conflict_info_[row][column].conflict_variables_ =
            segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
              conflicts_;
        }

        conflict_table_[row][column] |=
          segment_graph_.graph_[segment_id_to_vertex_descriptor(row_seg)].
              in_conflict_with_all_other_segments;

      } else {

        conflict_table_[row][column] = data_conflict_between(
            segment_id_to_vertex_descriptor(row_seg),
            segment_id_to_vertex_descriptor(column_seg),
            row,
            column,
            row_inst,
            column_inst);

        conflict_table_[column][row] = conflict_table_[row][column];
      }
    }
  }
}

void
risc::sg::DataConflictTable::color_members_of_class(
    SgVariableSymbol *var_symbol,
    DataConflictAstAttributes::Status color)
{
  // we get the type of the variable
  SgInitializedName *var_decl_stmt = var_symbol->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return;
  }

  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  for(Rose_STL_Container<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {

      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {


        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute == NULL) {
          data_conflict_attribute = new DataConflictAstAttributes();
          data_conflict_attribute->status_ = color;
          var_symbol->setAttribute("data_conflict", data_conflict_attribute);

        } else {
          data_conflict_attribute->status_ = color;

        }
      }
    }
  }
}

void
risc::sg::DataConflictTable::uncolor_members_of_class(
    SgVariableSymbol *var_symbol)
{
  // we get the type of the variable
  SgInitializedName *var_decl_stmt = var_symbol->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return;
  }

  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  for(Rose_STL_Container<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {

      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        SgVariableSymbol *var_symbol =
          isSgVariableSymbol(var_decl->get_definition()->get_vardefn()->
                              search_for_symbol_from_symbol_table());

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {

          data_conflict_attribute->status_ = DataConflictAstAttributes::UNUSED;
        }
      }
    }
  }
}

bool
risc::sg::DataConflictTable::check_members_of_class(
    SgVariableSymbol *var_symbol,
    DataConflictAstAttributes::Status color)
{
  SgInitializedName *var_decl_stmt = var_symbol->get_declaration();
  SgType *type = var_decl_stmt->get_type();
  SgDeclarationStatement *type_decl_stmt = type->getAssociatedDeclaration();

  if(!type_decl_stmt) {
    return false;
  }

  // we get the type of the variable
  SgClassDeclaration *class_decl = isSgClassDeclaration(type_decl_stmt);

  if(!class_decl) {
    return false;
  }

  class_decl = isSgClassDeclaration(class_decl->get_definingDeclaration());

  SgClassDefinition *class_def = class_decl->get_definition();

  for(Rose_STL_Container<SgDeclarationStatement*>::iterator
      member_iter  = class_def->get_members().begin();
      member_iter != class_def->get_members().end();
      member_iter++) {

    SgVariableDeclaration *var_decl = isSgVariableDeclaration(*member_iter);
    if(var_decl) {
      if(!var_decl->get_decl_item(0)->get_storageModifier().isStatic()) {

        DataConflictAstAttributes *data_conflict_attribute =
          dynamic_cast<DataConflictAstAttributes*>(
              var_symbol->getAttribute("data_conflict"));

        if(data_conflict_attribute != NULL) {

          if(data_conflict_attribute->status_ == color) {
            return true;
          }
        }
      }
    }
  }
  return false;
}


bool
risc::sg::DataConflictTable::data_conflict_between(VertexDescriptor first,
    VertexDescriptor second, int row, int column, int inst_id_first,
    int inst_id_second)
{
  if(segment_graph_.graph_[second].in_conflict_with_all_other_segments
      || segment_graph_.graph_[first].in_conflict_with_all_other_segments) {

    if(segment_graph_.graph_[first].in_conflict_with_all_other_segments) {

      conflict_info_[row][column].conflict_variables_.insert(
          segment_graph_.graph_[first].conflicts_.begin(),
          segment_graph_.graph_[first].conflicts_.end());

      conflict_info_[column][row].conflict_variables_.insert(
          segment_graph_.graph_[first].conflicts_.begin(),
          segment_graph_.graph_[first].conflicts_.end());
    }

    if(segment_graph_.graph_[second].in_conflict_with_all_other_segments) {

      conflict_info_[row][column].conflict_variables_.insert(
          segment_graph_.graph_[second].conflicts_.begin(),
          segment_graph_.graph_[second].conflicts_.end());

      conflict_info_[column][row].conflict_variables_.insert(
          segment_graph_.graph_[second].conflicts_.begin(),
          segment_graph_.graph_[second].conflicts_.end());
    }

    return true;
  }

  std::set<SgNode*> conflicts;

  bool conflict = false;

  for(std::set<SgVariableSymbol*>::iterator
      read_iter  = segment_graph_.graph_[first].read_variables_.begin();
      read_iter != segment_graph_.graph_[first].read_variables_.end();
      read_iter++) {

    SgVariableSymbol *read_symbol
      = get_mapped_variable(*read_iter, inst_id_first);

    if((tools::is_local_variable(read_symbol) &&
        !tools::is_static_variable(read_symbol)) ||
        tools::is_function_parameter_variable(read_symbol)) {
      continue;
    }

    DataConflictAstAttributes *data_conflict_attribute =
      dynamic_cast<DataConflictAstAttributes*>(
          (read_symbol)->getAttribute("data_conflict"));

    if(data_conflict_attribute == NULL) {
      data_conflict_attribute = new DataConflictAstAttributes();
      data_conflict_attribute->status_ = DataConflictAstAttributes::READ;
      (read_symbol)->setAttribute("data_conflict", data_conflict_attribute);

    } else {
      data_conflict_attribute->status_ = DataConflictAstAttributes::READ;
    }

    // In the case that we are reading a non pod data type we have to mark
    // all class/struct members as read
    color_members_of_class(read_symbol, DataConflictAstAttributes::READ);
  }

  for(std::set<SgVariableSymbol*>::iterator
      write_iter  = segment_graph_.graph_[first].write_variables_.begin();
      write_iter != segment_graph_.graph_[first].write_variables_.end();
      write_iter++) {

    SgVariableSymbol *write_symbol
      = get_mapped_variable(*write_iter, inst_id_first);

    if((tools::is_local_variable(write_symbol) &&
        !tools::is_static_variable(write_symbol)) ||
        tools::is_function_parameter_variable(write_symbol)) {
      continue;
    }

    DataConflictAstAttributes *data_conflict_attribute =
      dynamic_cast<DataConflictAstAttributes*>(
          (write_symbol)->getAttribute("data_conflict"));

    if(data_conflict_attribute == NULL) {
      data_conflict_attribute = new DataConflictAstAttributes();
      data_conflict_attribute->status_ = DataConflictAstAttributes::WRITE;
      (write_symbol)->setAttribute("data_conflict", data_conflict_attribute);

    } else {
      data_conflict_attribute->status_ = DataConflictAstAttributes::WRITE;
    }

    // In the case that we are reading a non pod data type we have to mark
    // all class/struct members as write
    color_members_of_class(write_symbol, DataConflictAstAttributes::WRITE);
  }

  for(std::set<SgVariableSymbol*>::iterator
      read_iter  = segment_graph_.graph_[second].read_variables_.begin();
      read_iter != segment_graph_.graph_[second].read_variables_.end();
      read_iter++) {

    SgVariableSymbol *read_symbol
      = get_mapped_variable(*read_iter, inst_id_second);

    if((tools::is_local_variable(read_symbol) &&
        !tools::is_static_variable(read_symbol)) ||
        tools::is_function_parameter_variable(read_symbol)) {
      continue;
    }

    DataConflictAstAttributes *data_conflict_attribute =
      dynamic_cast<DataConflictAstAttributes*>(
          (read_symbol)->getAttribute("data_conflict"));

    if(data_conflict_attribute != NULL) {

      if(data_conflict_attribute->status_ == DataConflictAstAttributes::WRITE) {

        if(tools::is_reference_variable(*read_iter)){
          conflict = true;

        } else if(tools::is_member_variable(read_symbol)) {
          if(inst_id_first == inst_id_second) {
            conflict = true;
          }

        } else {
          conflict = true;
        }

        if(debugging_) {
          conflicts.insert(read_symbol);
        } else {
          break;
        }
      }
    }

    conflict |= check_members_of_class(read_symbol,
                    DataConflictAstAttributes::WRITE);
  }

  for(std::set<SgVariableSymbol*>::iterator
      write_iter  = segment_graph_.graph_[second].write_variables_.begin();
      write_iter != segment_graph_.graph_[second].write_variables_.end();
      write_iter++) {


    SgVariableSymbol *write_symbol
      = get_mapped_variable(*write_iter, inst_id_second);

    if((tools::is_local_variable(write_symbol) &&
        !tools::is_static_variable(write_symbol)) ||
        tools::is_function_parameter_variable(write_symbol)) {
      continue;
    }

    DataConflictAstAttributes *data_conflict_attribute =
      dynamic_cast<DataConflictAstAttributes*>(
          (write_symbol)->getAttribute("data_conflict"));

    if(data_conflict_attribute != NULL) {

      if(data_conflict_attribute->status_ !=
          DataConflictAstAttributes::UNUSED) {

        if(tools::is_reference_variable(*write_iter)){
          conflict = true;

        } else if(tools::is_member_variable(write_symbol)) {
          if(inst_id_first == inst_id_second) {
            conflict = true;
          }

        } else {
          conflict = true;

        }

        if(debugging_) {
          conflicts.insert(write_symbol);
        } else {
          break;
        }
      }
    }

    conflict |= check_members_of_class(write_symbol,
                    DataConflictAstAttributes::WRITE);
    conflict |= check_members_of_class(write_symbol,
                    DataConflictAstAttributes::READ);
  }

  for(std::set<SgVariableSymbol*>::iterator
      read_iter  = segment_graph_.graph_[first].read_variables_.begin();
      read_iter != segment_graph_.graph_[first].read_variables_.end();
      read_iter++) {

    SgVariableSymbol *read_symbol
      = get_mapped_variable(*read_iter, inst_id_first);

    if((tools::is_local_variable(read_symbol) &&
        !tools::is_static_variable(read_symbol)) ||
        tools::is_function_parameter_variable(read_symbol)) {
      continue;
    }

    DataConflictAstAttributes *data_conflict_attribute =
      dynamic_cast<DataConflictAstAttributes*>(
          (read_symbol)->getAttribute("data_conflict"));
    data_conflict_attribute->status_ = DataConflictAstAttributes::UNUSED;

    uncolor_members_of_class(read_symbol);
  }

  for(std::set<SgVariableSymbol*>::iterator
      write_iter  = segment_graph_.graph_[first].write_variables_.begin();
      write_iter != segment_graph_.graph_[first].write_variables_.end();
      write_iter++) {

    SgVariableSymbol *write_symbol
      = get_mapped_variable(*write_iter, inst_id_first);

    if((tools::is_local_variable(write_symbol) &&
        !tools::is_static_variable(write_symbol)) ||
        tools::is_function_parameter_variable(write_symbol)) {
      continue;
    }

    DataConflictAstAttributes *data_conflict_attribute =
      dynamic_cast<DataConflictAstAttributes*>(
          (write_symbol)->getAttribute("data_conflict"));
    data_conflict_attribute->status_ = DataConflictAstAttributes::UNUSED;

    uncolor_members_of_class(write_symbol);
  }

  if(debugging_) {
    if(conflict) {

      for(std::set<SgNode*>::iterator
          iter  = conflicts.begin();
          iter != conflicts.end();
          iter++) {
        SgVariableSymbol *var_sym = isSgVariableSymbol(*iter);
      }
    }

    conflict_info_[row][column].conflict_variables_ = conflicts;
    conflict_info_[column][row].conflict_variables_ = conflicts;
  }

  return conflict;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
