#include "event_conflict_table.h"

#include <list>

#include "../tools/class_is_derived_from.h"
#include "../internal_representation/instance.h"
#include "../internal_representation/design.h"

std::list<SgVariableSymbol*>
get_all_and_events(SgSymbol *sc_event_and_list)
{
  std::list<SgVariableSymbol*> result;

  SgClassDefinition *class_def
    = SageInterface::getEnclosingClassDefinition(sc_event_and_list);

  std::vector<SgNode*> func_defs
    = NodeQuery::querySubTree(class_def, V_SgFunctionDefinition);

  SgFunctionDefinition *before_end_of_elaboration = NULL;

  for(std::vector<SgNode*>::iterator
      iter  = func_defs.begin();
      iter != func_defs.end();
      iter++) {

    SgFunctionDefinition *func_def = isSgFunctionDefinition(*iter);
    if(SageInterface::get_name(func_def) == "before_end_of_elaboration") {
      before_end_of_elaboration = func_def;
    }
  }

  for(SgStatementPtrList::iterator
      iter  = before_end_of_elaboration->get_body()->get_statements().begin();
      iter != before_end_of_elaboration->get_body()->get_statements().end();
      iter++) {

    SgExprStatement *expr_stmt = isSgExprStatement(*iter);

    if(!expr_stmt) {
      continue;
    }

    SgExpression *expr = isSgExpression(expr_stmt->get_expression());

    if(!expr) {
      continue;
    }

    SgFunctionCallExp *func_call_exp = isSgFunctionCallExp(expr);
    if(func_call_exp) {

      if(SageInterface::get_name(func_call_exp).find("&=")) {

        SgVarRefExp *assigned_var =
          isSgVarRefExp(
              isSgArrowExp(
                  isSgDotExp(
                      func_call_exp->get_function())
                  ->get_lhs_operand())
              ->get_rhs_operand());

        if(assigned_var) {

          if(assigned_var->get_symbol() == sc_event_and_list) {

            SgExprListExp *expr_list = func_call_exp->get_args();

            SgVarRefExp *var
              = isSgVarRefExp(
                    isSgArrowExp(
                        expr_list->get_expressions().at(0))
                    ->get_rhs_operand());

            result.push_back(var->get_symbol());
          }
        }
      }
    }

  }

  return result;
}

risc::sg::EventConflictTable::EventConflictTable(
    SegmentGraph &graph,
    PathInstanceMapper &path_instance_mapper,
    bool debugging):
  ConflictTable(graph, path_instance_mapper, debugging)
{
  determine_conflict_table();
}

void
risc::sg::EventConflictTable::determine_conflict_table()
{
  for(int column=0; column < size_of_conflict_table_; column++) {

    // Here, we are getting the segment id and instance id with the associated
    // column index 'column'.
    std::pair<int, int> column_id_and_instance =
      index_to_segment_and_instance_id(column);

    // Next, we have to get the boundary statement
    const SgNode *boundary
      = segment_graph_.graph_[
            segment_id_to_vertex_descriptor(column_id_and_instance.first)].
                boundary_node_;

    const SgStatement *stmt = isSgStatement(boundary);
    const SgExprStatement *expr_stmt = isSgExprStatement(stmt);


    std::list<SgVariableSymbol*> waiting_symbols;
    //SgSymbol *waiting_symbol = NULL;

    // Here, we have to check what kind of boundary statement it is. If it is a
    // wait function call, we have to analyze the arguments. So, if we have
    // exactly one argument it could be an event or an event list. We will check
    // if the symbol is of the type event.
    if(expr_stmt) {

      const SgFunctionCallExp *func_call
        = isSgFunctionCallExp(expr_stmt->get_expression());

      if(func_call) {

        if(SageInterface::get_name(
               func_call->getAssociatedFunctionDeclaration()) == "wait") {

          // At this point, we know that the boundary statement is a function
          // call of wait.
          // Next we have to figure out what are the arguments

          // In case we have an immediate notification.
          // We can not support this one
          if(func_call->get_args()->get_expressions().size() == 0) {

            std::cout << "Line: "
                      << func_call->get_file_info()->get_line()
                      << std::endl;
            assert(0 && "EventConflictTable::determine_conflict_table:" &&
                " No support of immediate notification");

          } else if(func_call->get_args()->get_expressions().size() == 1) {

            // A sc_zero_time will wait for one delta count. We don't have to
            // set the wait symbol
            if(func_call->get_args()->get_expressions().at(0)->unparseToString()
                == "SC_ZERO_TIME") {

            } else {

              // Here we are getting the variable for which the wait function
              // call is waiting
              SgVarRefExp* waiting_event
                = isSgVarRefExp(isSgArrowExp(func_call->get_args()->
                      get_expressions().at(0))->get_rhs_operand());

              // If it is a reference type, we use the InstanceTree to get
              // the real mapped variable
              if(isSgReferenceType(waiting_event->get_type())) {

                // In the following, we have to determine the real mapped
                // variable behind this reference.
                // First, we are searching for all potential occurrences of this
                // variable in the design. Afterwards, we can filter out through
                // the instance id and the PathInstanceMapper the associated
                // mapped variable
                SgVariableDefinition *var_def
                  = isSgVariableDefinition(waiting_event->get_symbol()
                       ->get_declaration()->get_definition());

                std::list<InstanceTree> potential_instances
                  = risc::get_design()->get_root().find_all_instances_of(
                                                       var_def);

                InstanceTree waiting_event_tree
                  = path_instance_mapper_.get_instance_tree(
                      column_id_and_instance.second, potential_instances);

                std::list<risc::InstanceTree> temp
                  = waiting_event_tree.find_all_instances_of(var_def);


                waiting_symbols.push_back
                  (temp.front().get_real_mapped_variable().get_object_symbol());


              } else if(isSgPointerType(waiting_event->get_type())) {

                assert(0 && "Event pointers are not supported");

              } else if(risc::tools::is_sc_event(waiting_event)) {

                // If it a event variable, we can set the symbol straight
                waiting_symbols.push_back(waiting_event->get_symbol());

              } else {

                waiting_symbols
                  = get_all_and_events(waiting_event->get_symbol());

                //assert(0 && "EventConflictTable::determine_conflict_table:" &&
                //" No support of sc_event_and/or_list.");
              }
            }
          }
        }
      }
    }

    // At this point, we have the waiting symbol of the wait function call.
    // We have to figure out, if this symbol is notified in each of the
    // segments.
    for(int row=0; row < size_of_conflict_table_; row++) {

      conflict_table_[row][column] = false;

      // Here, we are getting the segment id and instance id with the associated
      // row index 'row'.
      std::pair<int, int> row_id_and_instance =
        index_to_segment_and_instance_id(row);

      // We are iterating over all expressions for each segment. For an
      // individual segment we have to check if there are any nofitications of
      // events
      for(Segment::Expressions::iterator
          iter  = segment_graph_.graph_[
                      segment_id_to_vertex_descriptor(
                        row_id_and_instance.first)].expressions_.begin();
          iter != segment_graph_.graph_[
                      segment_id_to_vertex_descriptor(
                        row_id_and_instance.first)].expressions_.end();
          iter++) {

        SgExprStatement *expr_stmt = isSgExprStatement(*iter);
        SgFunctionCallExp *func_call = isSgFunctionCallExp(*iter);

        if(func_call) {

          SgFunctionDeclaration *func_decl
            = func_call->getAssociatedFunctionDeclaration();

          if(func_decl) {

            if(SageInterface::get_name(func_decl) == "notify") {

              // In case we have an immediate notification.
              // We can not support this one
              if(func_call->get_args()->get_expressions().size() == 0) {

                assert(0 && "EventConflictTable::determine_conflict_table:" &&
                    " No support of immediate notification");

              } else if(func_call->get_args()->get_expressions().size() == 1) {

                // At this point we know that it is a event notification
                SgExpression *lhs = isSgDotExp(
                                        func_call->get_function())->
                                            get_lhs_operand();

                SgVarRefExp *notify_event = isSgVarRefExp(isSgArrowExp(lhs)->
                                           get_rhs_operand());

                SgVariableSymbol *notify_symbol = NULL;

                if(risc::tools::is_sc_event(notify_event)) {

                  // If it a event variable, we can set the symbol straight
                  notify_symbol = notify_event->get_symbol();

                } else if(isSgPointerType(notify_event->get_type())) {

                  assert(0 && "Event pointers are not supported");

                } else if(isSgReferenceType(notify_event->get_type())) {
                  // In the following, we have to determine the real mapped
                  // variable behind this reference.
                  // First, we are searching for all potential occurrences of
                  // this variable in the design. Afterwards, we can filter out
                  // through the instance id and the PathInstanceMapper the
                  // associated mapped variable

                  SgVariableDefinition *var_def
                    =isSgVariableDefinition(notify_event->get_symbol()
                         ->get_declaration()->get_definition());

                  std::list<InstanceTree> potential_instances
                    = risc::get_design()->get_root().find_all_instances_of(
                                                         var_def);

                  InstanceTree notify_event_tree
                    = path_instance_mapper_.get_instance_tree(
                        row_id_and_instance.second, potential_instances);

                  std::list<risc::InstanceTree> temp
                    = notify_event_tree.find_all_instances_of(var_def);

                  notify_symbol
                    = temp.front().get_real_mapped_variable().
                                     get_object_symbol();
                }


                std::list<SgVariableSymbol*>::iterator conflict
                  = std::find(
                      waiting_symbols.begin(),
                      waiting_symbols.end(),
                      notify_symbol);


                //if(notify_symbol == waiting_symbol) {
                if(conflict != waiting_symbols.end()) {

                  conflict_table_[row][column] = true;
                  if(debugging_) {
                    conflict_info_[row][column].conflict_variables_.insert(
                                                    notify_event->get_symbol());
                  }
                }
              } else {
                assert(0 && "Unsuppported case");
              }
            }
          }
        }
      }
    }
  }
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
