#include "cthread_analyzer.h"

#include <boost/range/algorithm/set_algorithm.hpp>
#include "../internal_representation/design.h"
#include "read_write_analysis.h"
#include "../tools/file_classifier.h"

SgFunctionCallExp*
risc::sa::find_first_wait_stmt(SgFunctionDefinition *func_def)
{
  SgStatementPtrList statements = func_def->get_body()->get_statements();

  for(SgStatementPtrList::iterator
      iter = statements.begin();
      iter != statements.end();
      iter++) {

    SgExprStatement *expr_stmt = isSgExprStatement(*iter);
    if(expr_stmt) {

      SgFunctionCallExp *func_call =
        isSgFunctionCallExp(expr_stmt->get_expression());
      if(func_call) {

        for(std::set<SgFunctionDefinition*>::iterator
            iter_wait  = risc::get_design()->systemc_definitions_->
                                                 sc_wait_funcs_.begin();
            iter_wait != risc::get_design()->systemc_definitions_->
                                                 sc_wait_funcs_.end();
            iter_wait++) {

          if((*iter_wait)->get_declaration()->
                 search_for_symbol_from_symbol_table()
              == func_call->getAssociatedFunctionSymbol()) {

            return func_call;
          }
        }
      }
    }
  }

  return NULL;
}

std::set<SgFunctionCallExp*>
risc::sa::traverse_for_non_nested_func_calls(SgFunctionDefinition *func_def)
{
  std::set<SgFunctionCallExp*> func_calls;
  SgFunctionCallExp* first_wait_call = find_first_wait_stmt(func_def);

  SgStatementPtrList statements = func_def->get_body()->get_statements();
  for(SgStatementPtrList::iterator
      iter = statements.begin();
      iter != statements.end();
      iter++) {

    SgExprStatement *expr_stmt = isSgExprStatement(*iter);
    if(expr_stmt) {

      SgFunctionCallExp *func_call =
        isSgFunctionCallExp(expr_stmt->get_expression());
      if(func_call) {

        // the loop stops if the are at the position of the first wait stmt
        if(func_call == first_wait_call)
          break;

        func_calls.insert(func_call);
      }
    }
  }

  return func_calls;
}



std::set<SgFunctionDefinition*>
risc::sa::get_called_functions_and_expressions(
    SgFunctionDefinition *root_function,
    std::vector<SgExpression*> &expressions)
{
  std::set<SgFunctionDefinition*> result;
  SgStatementPtrList statements = root_function->get_body()->get_statements();

  for(std::vector<SgStatement*>::iterator
      stmt_iter  = statements.begin();
      stmt_iter != statements.end();
      stmt_iter++) {

    SgExprStatement *expr_stmt = isSgExprStatement(*stmt_iter );

    if(!isSgScopeStatement(*stmt_iter)) {


      Rose_STL_Container<SgNode*> temp_expressions =
        NodeQuery::querySubTree(*stmt_iter , V_SgExpression);

      for(Rose_STL_Container<SgNode*>::iterator
          iter_expr  = temp_expressions.begin();
          iter_expr != temp_expressions.end();
          iter_expr++) {
        expressions.push_back(isSgExpression(*iter_expr));
      }
    }

    if(expr_stmt) {

      SgFunctionCallExp *func_call =
        isSgFunctionCallExp(expr_stmt->get_expression());

      if(func_call) {

        SgFunctionDeclaration *func_decl =
          func_call->getAssociatedFunctionDeclaration();
        func_decl =
          isSgFunctionDeclaration(func_decl->get_definingDeclaration());

        if(func_decl != NULL) {
          SgFunctionDefinition *func_def = func_decl->get_definition();

          if(risc::tools::get_library(func_decl->get_file_info()) == "User") {

            result.insert(func_def);
          }
        }
      }
    }
  }

  return result;
}


void
risc::sa::get_recursivly_nested_expressions_in_function(
    SgFunctionDefinition *root_function,
    std::vector<SgExpression*> &expressions)
{
  std::set<SgFunctionDefinition*> analyzed_functions;
  std::set<SgFunctionDefinition*> unanalyzed_functions
    = risc::sa::get_called_functions_and_expressions(root_function,
                     expressions);

  while(!unanalyzed_functions.empty()) {

    std::set<SgFunctionDefinition*>::iterator iter =
      unanalyzed_functions.begin();
    SgFunctionDefinition *func_def = *iter;
    unanalyzed_functions.erase(iter);
    std::set<SgFunctionDefinition*> called_functions;
    boost::set_difference(
               get_called_functions_and_expressions(func_def, expressions),
               analyzed_functions,
               std::inserter(called_functions, called_functions.begin()));

    analyzed_functions.insert(func_def);
    unanalyzed_functions.insert(called_functions.begin(),
                             called_functions.end());
  }
}

void
risc::sa::collect_non_nested_expressions(SgFunctionDefinition *func_def,
    std::vector<SgExpression*> &before_wait,
    std::vector<SgExpression*> &after_wait)
{
  SgNode* first_wait_call = find_first_wait_stmt(func_def);

  SgExprStatement *first_wait_call_stmt =
    isSgExprStatement(first_wait_call->get_parent());

  SgStatementPtrList statements = func_def->get_body()->get_statements();

  SgStatementPtrList::iterator stmt_iter;

  for(stmt_iter = statements.begin();
      *stmt_iter != first_wait_call_stmt;
      stmt_iter++) {

    if(isSgScopeStatement(*stmt_iter))
      continue;

    // This considers all function nested assignments
    if(isSgExprStatement(*stmt_iter)) {

      SgExprStatement *expr_stmt = isSgExprStatement(*stmt_iter );
      SgFunctionCallExp *func_call =
        isSgFunctionCallExp(expr_stmt->get_expression());

      if(func_call) {

        SgFunctionDeclaration *func_decl =
          func_call->getAssociatedFunctionDeclaration();
        func_decl =
          isSgFunctionDeclaration(func_decl->get_definingDeclaration());

        if(func_decl != 0) {

          if(risc::tools::get_library(func_decl->get_file_info()) == "User") {

            SgFunctionDefinition *func_def = func_decl->get_definition();

            get_recursivly_nested_expressions_in_function(func_def,
                before_wait);
          }
        }
      }
    }

    Rose_STL_Container<SgNode*> expressions =
      NodeQuery::querySubTree(*stmt_iter , V_SgExpression);

    for(Rose_STL_Container<SgNode*>::iterator
        iter_expr  = expressions.begin();
        iter_expr != expressions.end();
        iter_expr++) {
      before_wait.push_back(isSgExpression(*iter_expr));
    }
  }

  for(++stmt_iter; stmt_iter != statements.end(); stmt_iter++) {

    Rose_STL_Container<SgNode*> expressions =
      NodeQuery::querySubTree(*stmt_iter , V_SgExpression);

    for(Rose_STL_Container<SgNode*>::iterator
        iter_expr  = expressions.begin();
        iter_expr != expressions.end();
        iter_expr++) {
      after_wait.push_back(isSgExpression(*iter_expr));
    }
  }
}


risc::sa::CThreadUninitializedVariablesSet
risc::sa::most_pessimistic_analysis(Module &module)
{
  CThreadUninitializedVariablesSet result;

  for(risc::CThreadVectorIter
      iter_cthreads  = module.cthreads_.begin();
      iter_cthreads != module.cthreads_.end();
      iter_cthreads++) {

    std::vector<SgExpression*> before_wait, after_wait;

    collect_non_nested_expressions((*iter_cthreads)->get_ast_node(),
        before_wait, after_wait);

    std::set<SgVariableSymbol*> read_only_var_before_wait,
      write_only_var_before_wait, read_only_var_after_wait,
      write_only_var_after_wait;

    std::vector<SgNode*> read_write_before_wait, read_write_after_wait;

    // analysis of the expression before the wait statement
    read_write_analysis(before_wait, read_only_var_before_wait,
      write_only_var_before_wait, read_write_before_wait);

    // analysis of the expression after the wait statement
    read_write_analysis(after_wait, read_only_var_after_wait,
      write_only_var_after_wait, read_write_after_wait);


    // delte the non member variables before the wait statement
    read_write_before_wait.erase(
        delete_if_is_not_member_variable(
            read_write_before_wait,
            module.variables_),
        read_write_before_wait.end());

    delete_if_is_not_member_symbol(
        write_only_var_before_wait,
        module.variables_);

    delete_if_is_not_member_symbol(
        read_only_var_before_wait,
        module.variables_);


    // delte the non member variables after the wait statement
    read_write_after_wait.erase(
        delete_if_is_not_member_variable(
            read_write_after_wait,
            module.variables_),
        read_write_after_wait.end());

    delete_if_is_not_member_symbol(
        write_only_var_after_wait,
        module.variables_);

    delete_if_is_not_member_symbol(
        read_only_var_after_wait,
        module.variables_);

    // Determining the set of written variables before the first wait statement
    std::set<SgVariableSymbol*> write_set_before_wait;

    for(std::vector<SgNode*>::iterator
        iter  = read_write_before_wait.begin();
        iter != read_write_before_wait.end();
        iter++) {

      SgVariableSymbol* symbol = NULL;

      if(isSgInitializedName(*iter)) {
        symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                     search_for_symbol_from_symbol_table());
      }

      if(isSgVarRefExp(*iter)) {
        symbol = isSgVarRefExp(*iter)->get_symbol();
      }

      write_set_before_wait.insert(symbol);
    }

    for(std::set<SgVariableSymbol*>::iterator
        iter  = write_only_var_before_wait.begin();
        iter != write_only_var_before_wait.end();
        iter++) {

      write_set_before_wait.insert(*iter);
    }

    // Determining the set of read variables after the first wait statement
    std::set<SgVariableSymbol*> read_set_after_wait;

    for(std::vector<SgNode*>::iterator
        iter  = read_write_after_wait.begin();
        iter != read_write_after_wait.end();
        iter++) {

      SgVariableSymbol* symbol = NULL;

      if(isSgInitializedName(*iter)) {
        symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                     search_for_symbol_from_symbol_table());
      }

      if(isSgVarRefExp(*iter)) {
        symbol = isSgVarRefExp(*iter)->get_symbol();
      }

      read_set_after_wait.insert(symbol);
    }

    for(std::set<SgVariableSymbol*>::iterator
        iter  = read_only_var_after_wait.begin();
        iter != read_only_var_after_wait.end();
        iter++) {

      read_set_after_wait.insert(*iter);
    }

    std::set<SgVariableSymbol*> potential_conflict;

    boost::set_difference(
               read_set_after_wait,
               write_set_before_wait,
               std::inserter(
                        potential_conflict,
                        potential_conflict.begin()));

    CThreadUninitializedVariables partial_result(*iter_cthreads,
                                      potential_conflict);
    //result.push_back(partial_result);
    result.insert(partial_result);
  }

  return result;
}


risc::sa::CThreadUninitializedVariablesSet
risc::sa::inter_thread_analysis(Module &module,
    CThreadVector *subset_of_cthreads)
{
  CThreadUninitializedVariablesSet result;

  std::vector<SgExpression*> before_wait_expr_all;
  std::vector<SgExpression*> after_wait_expr[module.cthreads_.size()];
  int cthread_counter = 0;

  // In the following loop we are intersted to get all the expressions from
  // all the CThreads
  for(risc::CThreadVectorIter
      iter_cthreads  = module.cthreads_.begin();
      iter_cthreads != module.cthreads_.end();
      iter_cthreads++, cthread_counter++) {

    std::vector<SgExpression*> before_wait_expr_temp;

    collect_non_nested_expressions((*iter_cthreads)->get_ast_node(),
        before_wait_expr_temp, after_wait_expr[cthread_counter]);

    before_wait_expr_all.insert(
        before_wait_expr_all.end(),
        before_wait_expr_temp.begin(),
        before_wait_expr_temp.end());
  }

  cthread_counter = 0;
  for(risc::CThreadVectorIter
      iter_cthreads  = module.cthreads_.begin();
      iter_cthreads != module.cthreads_.end();
      iter_cthreads++, cthread_counter++) {

    std::set<SgVariableSymbol*> read_only_var_before_wait,
      write_only_var_before_wait, read_only_var_after_wait,
      write_only_var_after_wait;

    std::vector<SgNode*> read_write_before_wait, read_write_after_wait;

    // analysis of the expression before the wait statement
    read_write_analysis(before_wait_expr_all, read_only_var_before_wait,
      write_only_var_before_wait, read_write_before_wait);

    // analysis of the expression after the wait statement
    read_write_analysis(after_wait_expr[cthread_counter],
        read_only_var_after_wait,
        write_only_var_after_wait,
        read_write_after_wait);

    // delte the non member variables before the wait statement
    read_write_before_wait.erase(
        delete_if_is_not_member_variable(
            read_write_before_wait,
            module.variables_),
        read_write_before_wait.end());

    delete_if_is_not_member_symbol(
        write_only_var_before_wait,
        module.variables_);

    delete_if_is_not_member_symbol(
        read_only_var_before_wait,
        module.variables_);


    // delte the non member variables after the wait statement
    read_write_after_wait.erase(
        delete_if_is_not_member_variable(
            read_write_after_wait,
            module.variables_),
        read_write_after_wait.end());

    delete_if_is_not_member_symbol(
        write_only_var_after_wait,
        module.variables_);

    delete_if_is_not_member_symbol(
        read_only_var_after_wait,
        module.variables_);


    // Determining the set of written variables before the first wait statement
    std::set<SgVariableSymbol*> write_set_before_wait;

    for(std::vector<SgNode*>::iterator
        iter  = read_write_before_wait.begin();
        iter != read_write_before_wait.end();
        iter++) {

      SgVariableSymbol* symbol = NULL;

      if(isSgInitializedName(*iter)) {
        symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                     search_for_symbol_from_symbol_table());
      }

      if(isSgVarRefExp(*iter)) {
        symbol = isSgVarRefExp(*iter)->get_symbol();
      }

      write_set_before_wait.insert(symbol);
    }

    for(std::set<SgVariableSymbol*>::iterator
        iter  = write_only_var_before_wait.begin();
        iter != write_only_var_before_wait.end();
        iter++) {

      write_set_before_wait.insert(*iter);
    }

    // Determining the set of read variables after the first wait statement
    std::set<SgVariableSymbol*> read_set_after_wait;

    for(std::vector<SgNode*>::iterator
        iter  = read_write_after_wait.begin();
        iter != read_write_after_wait.end();
        iter++) {

      SgVariableSymbol* symbol = NULL;

      if(isSgInitializedName(*iter)) {
        symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                     search_for_symbol_from_symbol_table());
      }

      if(isSgVarRefExp(*iter)) {
        symbol = isSgVarRefExp(*iter)->get_symbol();
      }

      read_set_after_wait.insert(symbol);
    }

    for(std::set<SgVariableSymbol*>::iterator
        iter  = read_only_var_after_wait.begin();
        iter != read_only_var_after_wait.end();
        iter++) {

      read_set_after_wait.insert(*iter);
    }

    std::set<SgVariableSymbol*> potential_conflict;

    boost::set_difference(
               read_set_after_wait,
               write_set_before_wait,
               std::inserter(
                        potential_conflict,
                        potential_conflict.begin()));

    CThreadUninitializedVariables partial_result(*iter_cthreads,
                                      potential_conflict);
    result.insert(partial_result);
  }

  return result;
}

risc::sa::CThreadResetSignalExpressionsVector
risc::sa::get_cthread_async_reset_signal_expressions(Module &module)
{
  ResetSignalSensitiveCThreadsSet reset_sensitive_cthreads;

  // Implement here!
  for(ResetSignalVector::iterator
      module_reset_signal  = module.async_reset_signals_.begin();
      module_reset_signal != module.async_reset_signals_.end();
      module_reset_signal++) {

    ResetSignal *reset_signal = *module_reset_signal;

    CThreadVector sensitive_threads;

    for(CThreadVectorIter
        iter_cthreads  = module.cthreads_.begin();
        iter_cthreads != module.cthreads_.end();
        iter_cthreads++) {

      CThread *cthread = *iter_cthreads;

      for(ResetSignalVector::iterator
          iter_reset_signal  = cthread->async_reset_signal_.begin();
          iter_reset_signal != cthread->async_reset_signal_.end();
          iter_reset_signal++) {

        ResetSignal *reset_signal_cthread = *iter_reset_signal;

        if(reset_signal_cthread->first == reset_signal->first &&
            reset_signal_cthread->second->get_value() ==
              reset_signal->second->get_value()) {

          sensitive_threads.push_back(cthread);
        }
      }
    }
    ResetSignalSensitiveCThreads temp(reset_signal, sensitive_threads);
    reset_sensitive_cthreads.insert(temp);
  }


  CThreadResetSignalExpressionsVector cthread_rest_signal_expressions;

  for(ResetSignalSensitiveCThreadsSet::iterator
      reset_cthread_iter  = reset_sensitive_cthreads.begin();
      reset_cthread_iter != reset_sensitive_cthreads.end();
      reset_cthread_iter++) {

    ResetSignalSensitiveCThreads reset_signal_sensitive_cthreads =
      *reset_cthread_iter;

    for(CThreadVectorIter
        iter_cthreads  = reset_signal_sensitive_cthreads.second.begin();
        iter_cthreads != reset_signal_sensitive_cthreads.second.end();
        iter_cthreads++) {

      CThread *cthread = *iter_cthreads;
      SgFunctionDefinition *func_def = cthread->get_ast_node();

      SgNode* first_wait_call = find_first_wait_stmt(func_def);

      SgExprStatement *first_wait_call_stmt =
        isSgExprStatement(first_wait_call->get_parent());

      SgStatementPtrList statements = func_def->get_body()->get_statements();

      SgStatementPtrList::iterator stmt_iter;

      for(stmt_iter = statements.begin();
          *stmt_iter != first_wait_call_stmt;
          stmt_iter++) {

        SgIfStmt *if_stmt = isSgIfStmt(*stmt_iter);
        if(if_stmt) {

          SgExprStatement *stmt_expr =
            isSgExprStatement(if_stmt->get_conditional());
          SgExpression *cond_expr = stmt_expr->get_expression();


          // getting the symbol of the reset signal variable
          ResetSignal *rs = reset_signal_sensitive_cthreads.first;
          SgVariableDeclaration *reset_var_decl =
            rs->first->get_ast_node()->get_declaration();

          SgVariableSymbol *reset_signal_symbol =
            isSgVariableSymbol(reset_var_decl->
                get_decl_item(0)->search_for_symbol_from_symbol_table());

          Rose_STL_Container<SgNode*> var_refs =
            NodeQuery::querySubTree(cond_expr, V_SgVarRefExp);

          if(var_refs.size() == 1) {

            SgVarRefExp *condition_var = isSgVarRefExp(var_refs.at(0));
            if(condition_var->get_symbol() == reset_signal_symbol) {

              Rose_STL_Container<SgNode*> expressions =
                NodeQuery::querySubTree(if_stmt->get_true_body(),
                               V_SgExpression);

              std::vector<SgExpression*> reset_expressions;
              for(Rose_STL_Container<SgNode*>::iterator
                  expr_iter  = expressions.begin();
                  expr_iter != expressions.end();
                  expr_iter++) {

                reset_expressions.push_back(isSgExpression(*expr_iter));
              }

              CThreadResetSignalPair cthread_reset(cthread,
                                         reset_signal_sensitive_cthreads.
                                             first);

              CThreadResetSignalExpressions temp_cthread_reste_expr(
                                                cthread_reset,
                                                reset_expressions);

              cthread_rest_signal_expressions.push_back(
                                                  temp_cthread_reste_expr);
            }
          }
        }
      }
    }
  }

  return cthread_rest_signal_expressions;
}


risc::sa::CThreadResetSignalExpressionsVector
risc::sa::get_cthread_reset_signal_expressions(Module &module)
{
  ResetSignalSensitiveCThreadsSet reset_sensitive_cthreads;

  // Implement here!
  for(ResetSignalVector::iterator
      module_reset_signal  = module.reset_signals_.begin();
      module_reset_signal != module.reset_signals_.end();
      module_reset_signal++) {

    ResetSignal *reset_signal = *module_reset_signal;

    CThreadVector sensitive_threads;

    for(CThreadVectorIter
        iter_cthreads  = module.cthreads_.begin();
        iter_cthreads != module.cthreads_.end();
        iter_cthreads++) {

      CThread *cthread = *iter_cthreads;

      for(ResetSignalVector::iterator
          iter_reset_signal  = cthread->reset_signal_.begin();
          iter_reset_signal != cthread->reset_signal_.end();
          iter_reset_signal++) {

        ResetSignal *reset_signal_cthread = *iter_reset_signal;

        if(reset_signal_cthread->first == reset_signal->first &&
            reset_signal_cthread->second->get_value()
               == reset_signal->second->get_value()) {

          sensitive_threads.push_back(cthread);
        }
      }
    }
    ResetSignalSensitiveCThreads temp(reset_signal, sensitive_threads);
    reset_sensitive_cthreads.insert(temp);
  }


  CThreadResetSignalExpressionsVector cthread_rest_signal_expressions;

  for(ResetSignalSensitiveCThreadsSet::iterator
      reset_cthread_iter  = reset_sensitive_cthreads.begin();
      reset_cthread_iter != reset_sensitive_cthreads.end();
      reset_cthread_iter++) {

    ResetSignalSensitiveCThreads reset_signal_sensitive_cthreads =
      *reset_cthread_iter;

    for(CThreadVectorIter
        iter_cthreads  = reset_signal_sensitive_cthreads.second.begin();
        iter_cthreads != reset_signal_sensitive_cthreads.second.end();
        iter_cthreads++) {

      CThread *cthread = *iter_cthreads;
      SgFunctionDefinition *func_def = cthread->get_ast_node();

      SgNode* first_wait_call = find_first_wait_stmt(func_def);

      SgExprStatement *first_wait_call_stmt =
        isSgExprStatement(first_wait_call->get_parent());

      SgStatementPtrList statements = func_def->get_body()->get_statements();

      SgStatementPtrList::iterator stmt_iter;

      for(stmt_iter = statements.begin();
          *stmt_iter != first_wait_call_stmt;
          stmt_iter++) {

        SgIfStmt *if_stmt = isSgIfStmt(*stmt_iter);
        if(if_stmt) {

          SgExprStatement *stmt_expr =
            isSgExprStatement(if_stmt->get_conditional());
          SgExpression *cond_expr = stmt_expr->get_expression();


          // getting the symbol of the reset signal variable
          ResetSignal *rs = reset_signal_sensitive_cthreads.first;
          SgVariableDeclaration *reset_var_decl =
            rs->first->get_ast_node()->get_declaration();

          SgVariableSymbol *reset_signal_symbol =
            isSgVariableSymbol(reset_var_decl->get_decl_item(0)->
                search_for_symbol_from_symbol_table());

          Rose_STL_Container<SgNode*> var_refs =
            NodeQuery::querySubTree(cond_expr, V_SgVarRefExp);

          if(var_refs.size() == 1) {

            SgVarRefExp *condition_var = isSgVarRefExp(var_refs.at(0));
            if(condition_var->get_symbol() == reset_signal_symbol) {

              Rose_STL_Container<SgNode*> expressions =
                NodeQuery::querySubTree(if_stmt->get_true_body(),
                               V_SgExpression);

              std::vector<SgExpression*> reset_expressions;
              for(Rose_STL_Container<SgNode*>::iterator
                  expr_iter  = expressions.begin();
                  expr_iter != expressions.end();
                  expr_iter++) {

                reset_expressions.push_back(isSgExpression(*expr_iter));
              }

              CThreadResetSignalPair cthread_reset(cthread,
                                         reset_signal_sensitive_cthreads.
                                             first);

              CThreadResetSignalExpressions temp_cthread_reste_expr(
                                                cthread_reset,
                                                reset_expressions);

              cthread_rest_signal_expressions.push_back(
                                                  temp_cthread_reste_expr);
            }
          }
        }
      }
    }
  }

  return cthread_rest_signal_expressions;
}

risc::sa::ResetSignalUninitializedVariablesSet
risc::sa::inter_async_reset_cthread_analysis(Module &module)
{
  ResetSignalUninitializedVariablesSet result;

  CThreadResetSignalExpressionsVector cthread_async_reset_expressions =
    get_cthread_async_reset_signal_expressions(module);

  ResetSignalSensitiveCThreadsSet async_reset_sensitive_cthreads;
  for(ResetSignalVector::iterator
      reset_signal_iter  = module.async_reset_signals_.begin();
      reset_signal_iter != module.async_reset_signals_.end();
      reset_signal_iter++) {

    CThreadVector cthread_reset;
    for(CThreadVector::iterator
        cthread_iter  = module.cthreads_.begin();
        cthread_iter != module.cthreads_.end();
        cthread_iter++) {

      for(ResetSignalVector::iterator
          cthread_reset_iter  = (*cthread_iter)->async_reset_signal_.begin();
          cthread_reset_iter != (*cthread_iter)->async_reset_signal_.end();
          cthread_reset_iter ++) {

        if(*reset_signal_iter == *cthread_reset_iter) {

          cthread_reset.push_back(*cthread_iter);
        }
      }
    }
    async_reset_sensitive_cthreads.insert(
                                       ResetSignalSensitiveCThreads(
                                           *reset_signal_iter, cthread_reset));
  }

  for(ResetSignalSensitiveCThreadsSet::iterator
      reset_sensitve_cthreads_iter  = async_reset_sensitive_cthreads.begin();
      reset_sensitve_cthreads_iter != async_reset_sensitive_cthreads.end();
      reset_sensitve_cthreads_iter++) {

    ResetSignalSensitiveCThreads reset_cthreads = *reset_sensitve_cthreads_iter;

    std::vector<SgExpression*> before_wait_expr_all;
    std::vector<SgExpression*> after_wait_expr[module.cthreads_.size()];
    int cthread_counter = 0;

    // In the following loop we are intersted to get all the expressions from
    // all the CThreads
    for(risc::CThreadVector::iterator
        iter_cthreads  = reset_cthreads.second.begin();
        iter_cthreads != reset_cthreads.second.end();
        iter_cthreads++, cthread_counter++) {

      std::vector<SgExpression*> before_wait_expr_temp;

      collect_non_nested_expressions((*iter_cthreads)->get_ast_node(),
          before_wait_expr_temp, after_wait_expr[cthread_counter]);

      // here we are adding the nested reset expressions
      for(CThreadResetSignalExpressionsVector::iterator
          reset_expressions  = cthread_async_reset_expressions.begin();
          reset_expressions != cthread_async_reset_expressions.end();
          reset_expressions ++) {

        CThreadResetSignalExpressions cthread_reset_signal_expressions =
          *reset_expressions;

        if(cthread_reset_signal_expressions.first.first == *iter_cthreads &&
            cthread_reset_signal_expressions.first.second
               == reset_cthreads.first) {

          before_wait_expr_all.insert(
              before_wait_expr_all.end(),
              cthread_reset_signal_expressions.second.begin(),
              cthread_reset_signal_expressions.second.end());
        }
      }

      before_wait_expr_all.insert(
          before_wait_expr_all.end(),
          before_wait_expr_temp.begin(),
          before_wait_expr_temp.end());
    }

    std::set<SgVariableSymbol*> potential_conflict;
    cthread_counter = 0;
    for(risc::CThreadVectorIter
        iter_cthreads  = module.cthreads_.begin();
        iter_cthreads != module.cthreads_.end();
        iter_cthreads++, cthread_counter++) {

      std::set<SgVariableSymbol*> read_only_var_before_wait,
        write_only_var_before_wait, read_only_var_after_wait,
        write_only_var_after_wait;

      std::vector<SgNode*> read_write_before_wait, read_write_after_wait;

      // analysis of the expression before the wait statement
      read_write_analysis(before_wait_expr_all, read_only_var_before_wait,
        write_only_var_before_wait, read_write_before_wait);

      // analysis of the expression after the wait statement
      read_write_analysis(after_wait_expr[cthread_counter],
          read_only_var_after_wait,
          write_only_var_after_wait,
          read_write_after_wait);


      // delte the non member variables before the wait statement
      read_write_before_wait.erase(
          delete_if_is_not_member_variable(
              read_write_before_wait,
              module.variables_),
          read_write_before_wait.end());

      delete_if_is_not_member_symbol(
          write_only_var_before_wait,
          module.variables_);

      delete_if_is_not_member_symbol(
          read_only_var_before_wait,
          module.variables_);


      // delte the non member variables after the wait statement
      read_write_after_wait.erase(
          delete_if_is_not_member_variable(
              read_write_after_wait,
              module.variables_),
          read_write_after_wait.end());

      delete_if_is_not_member_symbol(
          write_only_var_after_wait,
          module.variables_);

      delete_if_is_not_member_symbol(
          read_only_var_after_wait,
          module.variables_);


      // Determining the set of written variables before the first wait
      // statement
      std::set<SgVariableSymbol*> write_set_before_wait;

      for(std::vector<SgNode*>::iterator
          iter  = read_write_before_wait.begin();
          iter != read_write_before_wait.end();
          iter++) {

        SgVariableSymbol* symbol = NULL;

        if(isSgInitializedName(*iter)) {
          symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                       search_for_symbol_from_symbol_table());
        }

        if(isSgVarRefExp(*iter)) {
          symbol = isSgVarRefExp(*iter)->get_symbol();
        }

        write_set_before_wait.insert(symbol);
      }

      for(std::set<SgVariableSymbol*>::iterator
          iter  = write_only_var_before_wait.begin();
          iter != write_only_var_before_wait.end();
          iter++) {

        write_set_before_wait.insert(*iter);
      }

      // Determining the set of read variables after the first wait statement
      std::set<SgVariableSymbol*> read_set_after_wait;

      for(std::vector<SgNode*>::iterator
          iter  = read_write_after_wait.begin();
          iter != read_write_after_wait.end();
          iter++) {

        SgVariableSymbol* symbol = NULL;

        if(isSgInitializedName(*iter)) {
          symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                       search_for_symbol_from_symbol_table());
        }

        if(isSgVarRefExp(*iter)) {
          symbol = isSgVarRefExp(*iter)->get_symbol();
        }

        read_set_after_wait.insert(symbol);
      }

      for(std::set<SgVariableSymbol*>::iterator
          iter  = read_only_var_after_wait.begin();
          iter != read_only_var_after_wait.end();
          iter++) {

        read_set_after_wait.insert(*iter);
      }


      boost::set_difference(
          read_set_after_wait,
          write_set_before_wait,
          std::inserter(
              potential_conflict,
              potential_conflict.begin()));
    }

    ResetSignalUninitializedVariables partial_result(reset_cthreads.first,
                                          potential_conflict);
    result.insert(partial_result);
  }

  return result;
}

risc::sa::ResetSignalUninitializedVariablesSet
risc::sa::inter_reset_cthread_analysis(Module &module)
{
  ResetSignalUninitializedVariablesSet result;

  CThreadResetSignalExpressionsVector cthread_reset_expressions =
      get_cthread_reset_signal_expressions(module);


  ResetSignalSensitiveCThreadsSet reset_sensitive_cthreads;
  for(ResetSignalVector::iterator
      reset_signal_iter  = module.reset_signals_.begin();
      reset_signal_iter != module.reset_signals_.end();
      reset_signal_iter++) {

    CThreadVector cthread_reset;
    for(CThreadVector::iterator
        cthread_iter  = module.cthreads_.begin();
        cthread_iter != module.cthreads_.end();
        cthread_iter++) {

      for(ResetSignalVector::iterator
          cthread_reset_iter  = (*cthread_iter)->reset_signal_.begin();
          cthread_reset_iter != (*cthread_iter)->reset_signal_.end();
          cthread_reset_iter ++) {

        if(*reset_signal_iter == *cthread_reset_iter) {

          cthread_reset.push_back(*cthread_iter);
        }
      }
    }
    reset_sensitive_cthreads.insert(
                                 ResetSignalSensitiveCThreads(
                                     *reset_signal_iter, cthread_reset));
  }

  for(ResetSignalSensitiveCThreadsSet::iterator
      reset_sensitve_cthreads_iter  = reset_sensitive_cthreads.begin();
      reset_sensitve_cthreads_iter != reset_sensitive_cthreads.end();
      reset_sensitve_cthreads_iter++) {

    ResetSignalSensitiveCThreads reset_cthreads = *reset_sensitve_cthreads_iter;

    std::vector<SgExpression*> before_wait_expr_all;
    std::vector<SgExpression*> after_wait_expr[module.cthreads_.size()];
    int cthread_counter = 0;

    // In the following loop we are intersted to get all the expressions from
    // all the CThreads
    for(risc::CThreadVector::iterator
        iter_cthreads  = reset_cthreads.second.begin();
        iter_cthreads != reset_cthreads.second.end();
        iter_cthreads++, cthread_counter++) {

      std::vector<SgExpression*> before_wait_expr_temp;

      collect_non_nested_expressions((*iter_cthreads)->get_ast_node(),
          before_wait_expr_temp, after_wait_expr[cthread_counter]);

      // here we are adding the nested reset expressions
      for(CThreadResetSignalExpressionsVector::iterator
          reset_expressions  = cthread_reset_expressions.begin();
          reset_expressions != cthread_reset_expressions.end();
          reset_expressions ++) {

        CThreadResetSignalExpressions cthread_reset_signal_expressions =
          *reset_expressions;

        if(cthread_reset_signal_expressions.first.first == *iter_cthreads &&
            cthread_reset_signal_expressions.first.second
               == reset_cthreads.first) {

          before_wait_expr_all.insert(
              before_wait_expr_all.end(),
              cthread_reset_signal_expressions.second.begin(),
              cthread_reset_signal_expressions.second.end());
        }
      }

      before_wait_expr_all.insert(
          before_wait_expr_all.end(),
          before_wait_expr_temp.begin(),
          before_wait_expr_temp.end());
    }

    std::set<SgVariableSymbol*> potential_conflict;
    cthread_counter = 0;
    for(risc::CThreadVectorIter
        iter_cthreads  = module.cthreads_.begin();
        iter_cthreads != module.cthreads_.end();
        iter_cthreads++, cthread_counter++) {

      std::set<SgVariableSymbol*> read_only_var_before_wait,
        write_only_var_before_wait, read_only_var_after_wait,
        write_only_var_after_wait;

      std::vector<SgNode*> read_write_before_wait, read_write_after_wait;

      // analysis of the expression before the wait statement
      read_write_analysis(before_wait_expr_all, read_only_var_before_wait,
          write_only_var_before_wait, read_write_before_wait);

      // analysis of the expression after the wait statement
      read_write_analysis(after_wait_expr[cthread_counter],
          read_only_var_after_wait,
          write_only_var_after_wait,
          read_write_after_wait);


      // delte the non member variables before the wait statement
      read_write_before_wait.erase(
          delete_if_is_not_member_variable(
              read_write_before_wait,
              module.variables_),
          read_write_before_wait.end());

      delete_if_is_not_member_symbol(
          write_only_var_before_wait,
          module.variables_);

      delete_if_is_not_member_symbol(
          read_only_var_before_wait,
          module.variables_);


      // delte the non member variables after the wait statement
      read_write_after_wait.erase(
          delete_if_is_not_member_variable(
              read_write_after_wait,
              module.variables_),
          read_write_after_wait.end());

      delete_if_is_not_member_symbol(
          write_only_var_after_wait,
          module.variables_);

      delete_if_is_not_member_symbol(
          read_only_var_after_wait,
          module.variables_);


      // Determining the set of written variables before the first wait
      // statement
      std::set<SgVariableSymbol*> write_set_before_wait;

      for(std::vector<SgNode*>::iterator
          iter  = read_write_before_wait.begin();
          iter != read_write_before_wait.end();
          iter++) {

        SgVariableSymbol* symbol = NULL;

        if(isSgInitializedName(*iter)) {
          symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                       search_for_symbol_from_symbol_table());
        }

        if(isSgVarRefExp(*iter)) {
          symbol = isSgVarRefExp(*iter)->get_symbol();
        }

        write_set_before_wait.insert(symbol);
      }

      for(std::set<SgVariableSymbol*>::iterator
          iter  = write_only_var_before_wait.begin();
          iter != write_only_var_before_wait.end();
          iter++) {

        write_set_before_wait.insert(*iter);
      }

      // Determining the set of read variables after the first wait statement
      std::set<SgVariableSymbol*> read_set_after_wait;

      for(std::vector<SgNode*>::iterator
          iter  = read_write_after_wait.begin();
          iter != read_write_after_wait.end();
          iter++) {

        SgVariableSymbol* symbol = NULL;

        if(isSgInitializedName(*iter)) {
          symbol = isSgVariableSymbol(isSgInitializedName(*iter)->
                       search_for_symbol_from_symbol_table());
        }

        if(isSgVarRefExp(*iter)) {
          symbol = isSgVarRefExp(*iter)->get_symbol();
        }

        read_set_after_wait.insert(symbol);
      }

      for(std::set<SgVariableSymbol*>::iterator
          iter  = read_only_var_after_wait.begin();
          iter != read_only_var_after_wait.end();
          iter++) {

        read_set_after_wait.insert(*iter);
      }


      boost::set_difference(
          read_set_after_wait,
          write_set_before_wait,
          std::inserter(
                   potential_conflict,
                   potential_conflict.begin()));
    }

    ResetSignalUninitializedVariables partial_result(reset_cthreads.first,
                                          potential_conflict);
    result.insert(partial_result);
  }

  return result;
}

/* ex: set softtabstop=2 tabstop=2 shiftwidth=2 expandtab: */
